"""Represent fhir entity."""

from anvil.transformers.fhir import make_identifier
import logging
from anvil.transformers.fhir.disease_normalizer import ontology_text, disease_system
from anvil.transformers.fhir import CANONICAL
from anvil.transformers.fhir.patient import Patient
from anvil.transformers.fhir import make_id

logged_already = []


class DiseaseObservation:
    """Create fhir entity."""

    @staticmethod
    def slug(subject, disease):
        """Make id."""        
        return make_id(Patient.slug(subject), f"SNOMED:373573001/{disease}")


    @staticmethod
    def build_entity(subject, disease):
        """Create FHIR entity."""

        assert disease, f'Should have disease {subject}'
        if disease.startswith("PS"):
            disease = f"OMIM:{disease}"
        if ":" not in disease:
            disease = f"OMIM:{disease}"
        workspace_diseaseOntologyId = disease  # subject.workspace_diseaseOntologyId
        diseaseOntologyText = ontology_text.get(disease, None)
        diseaseOntologySystem = disease_system.get(disease.split(':')[0], None)
        workspace_diseaseOntologyId = disease.split(':')[1]
        slug = DiseaseObservation.slug(subject, workspace_diseaseOntologyId)

        if workspace_diseaseOntologyId and not diseaseOntologyText:
            if workspace_diseaseOntologyId not in logged_already:
                logging.getLogger(__name__).error(f'Need text "{workspace_diseaseOntologyId}"')
                logged_already.append(workspace_diseaseOntologyId)
        if not diseaseOntologySystem:
            if workspace_diseaseOntologyId not in logged_already:
                logging.getLogger(__name__).error(f"Should have system. {subject} {disease}")
            diseaseOntologySystem = "MISSING"

        entity = {
            "resourceType": "Observation",
            "id": slug,
            "meta": {
                "profile": [
                    f"http://{CANONICAL}/StructureDefinition/ncpi-phenotype"
                ]
            },
            "identifier": [
                {
                    "system": "urn:ncpi:unique-string",
                    "value": slug
                }
            ],
            "status": "final",
            "code": {
                "coding": [
                    {
                        "system": diseaseOntologySystem,
                        "code": workspace_diseaseOntologyId,
                        "display": diseaseOntologyText
                    }
                ],
                "text": f"{diseaseOntologyText}"
            },
            "subject": {
                "reference": f"Patient/{Patient.slug(subject)}"
            },
            "valueCodeableConcept": {
                "coding": [
                    {
                        "system": "http://snomed.info/sct",
                        "code": "373573001",
                        "display": "Clinical finding present (situation)"
                    }
                ],
                "text": "Phenotype Present"
            },
            "interpretation": [
                {
                    "coding": [
                        {
                            "system": "http://terminology.hl7.org/CodeSystem/v3-ObservationInterpretation",
                            "code": "POS",
                            "display": "Positive"
                        }
                    ],
                    "text": "Present"
                }
            ],
            "extension": []
        }
        if subject.age:
            entity['extension'].append(
                {
                    "url": f"http://{CANONICAL}/StructureDefinition/age-at-event",
                    "valueAge": {
                        "value": subject.age,
                        "unit": "y",
                        "system": "http://unitsofmeasure.org",
                        "code": "years"
                    }
                }
            )

        return entity



"""Gen3 auth helper class for use within the AnVIL (terra, gen3) environment."""

import logging
import requests
import base64

from datetime import datetime
from subprocess import Popen, PIPE
from requests.auth import AuthBase


TERRA_TOKEN_URL = "https://broad-bond-prod.appspot.com/api/link/v1/anvil/accesstoken"


class AnVILAuthError(Exception):
    """Reports any problem retrieving access token from terra."""

    pass


class Gen3TerraAuth(AuthBase):
    """Gen3 auth helper class for use with requests auth.

    Implements requests.auth.AuthBase in order to support JWT authentication.
        * Depends on gcloud cli and valid configuration
        * Validates with terra using token provided by 'gcloud auth print-access-token'
        * Queries terra endpoint for fence access_token.
        * Intercepts all calls to gen3 and inserts Bearer token
        * Automatically refreshes access tokens when they expire.

    Args:
        * terra_auth_url (str): URL of the terra endpoint Default: "https://broad-bond-prod.appspot.com/api/link/v1/anvil/accesstoken".
        * user_email (str): Optional, google id to pass to 'gcloud auth print-access-token' Default: None.

    Examples: ::

        # authenticate with terra
        from anvil.clients.gen3_auth import Gen3TerraAuth
        from gen3.submission import Gen3Submission

        gen3_endpoint = "https://gen3.theanvil.io"
        auth = Gen3TerraAuth(endpoint=gen3_endpoint)
        submission_client = Gen3Submission(gen3_endpoint, auth)
        query = '{project(first:0) {code,  subjects {submitter_id}, programs {name}  }}'
        results = submission_client.query(query)
        [p['code'] for p in results['data']['project']]
        >>> ['GTEx', '1000Genomes']

        # access terra api
        from anvil.terra.api import whoami
        whoami()
        >>> 'anvil.user@gmail.com'

    """

    def __init__(self, endpoint, terra_auth_url=TERRA_TOKEN_URL, user_email=None):
        """Initialize properties."""
        self._access_token = None
        self._terra_auth_url = terra_auth_url
        assert self._terra_auth_url, "MUST have _terra_auth_url"
        self._user_email = user_email
        self.endpoint = endpoint
        self._logger = logging.getLogger(__name__)

    def __call__(self, request):
        """Add authorization header to the request.

        This gets called by the python.requests package on outbound requests
        so that authentication can be added.

        Args:
            request (object): The incoming request object

        """
        self._logger.debug(f'__call__, {request.url} adding Authorization header')
        request.headers["Authorization"] = self._get_auth_value()
        request.register_hook("response", self._handle_401)
        return request

    def _handle_401(self, response, **kwargs):
        """Handle failed requests when authorization failed.

        This gets called after a failed request when an HTTP 401 error
        occurs. This then tries to refresh the access token in the event
        that it expired.

        Args:
            request (object): The failed request object

        """
        if not response.status_code == 401 and not response.status_code == 403:
            return response

        # Free the original connection
        response.content
        response.close()

        # copy the request to resend
        newreq = response.request.copy()

        self._access_token = None
        self._logger.debug("_handle_401, cleared _access_token, retrying with new token")

        newreq.headers["Authorization"] = self._get_auth_value()

        _response = response.connection.send(newreq, **kwargs)
        _response.history.append(response)
        _response.request = newreq

        return _response

    def _get_auth_value(self):
        """Return the Authorization header value for the request.

        This gets called when added the Authorization header to the request.
        This fetches the access token from the refresh token if the access token is missing.

        """
        if not self._access_token:
            try:
                # get the local access token using gcloud
                cmd = ['gcloud', 'auth', 'print-access-token']
                if self._user_email:
                    cmd.append(self._user_email)

                self._logger.debug(f"get gcloud_access_token {cmd}")
                p = Popen(cmd, stdout=PIPE, stderr=PIPE)
                gcloud_access_token, stderr = p.communicate()
                gcloud_access_token = gcloud_access_token.decode("utf-8").rstrip()
                assert len(gcloud_access_token) > 0, f'get gcloud_access_token MUST have an access token {stderr}'
                self._logger.debug(f"gcloud_access_token {gcloud_access_token}")
                # authenticate to terra, ask for fence/accesstoken
                headers = {'Authorization': f'Bearer {gcloud_access_token}'}
                r = requests.get(self._terra_auth_url, headers=headers)
                assert r.status_code == 200, f'MUST respond with 200 {self._terra_auth_url} {r.text}'
                self._logger.debug(r.text)
                terra_access_token = r.json()
                assert len(terra_access_token['token']) > 0, 'MUST have an access token'
                assert len(terra_access_token['expires_at']) > 0, 'MUST have an expires_at '

                expires_at = datetime.fromisoformat(terra_access_token['expires_at'])
                now = datetime.now()
                assert expires_at > now, 'expires_at MUST be in the future'

                self._access_token = terra_access_token['token']

                if self._logger.level == logging.DEBUG:
                    self._logger.debug(f'Terra access token expires in {str(expires_at - now)}')
                    self._logger.debug(self._access_token)
                    # add padding
                    self._logger.debug(base64.b64decode(self._access_token.split('.')[1] + "==="))

            except Exception as e:
                raise AnVILAuthError(
                    "Failed to authenticate to {}\n{}".format(self._terra_auth_url, str(e))
                )

        return "Bearer " + self._access_token



import os

DEFAULT_OUTPUT_PATH = os.environ.get('OUTPUT_PATH','./DATA')

DEFAULT_GEN3_CREDENTIALS_PATH = os.path.expanduser('~/.gen3/credentials.json')

DEFAULT_GOOGLE_PROJECT = os.environ.get('GOOGLE_PROJECT', None)


"""Client for reconciling terra workspace with the Google Cloud Storage API, dbGap and Gen3."""

import logging
from attrdict import AttrDict
from google.cloud import storage
from collections import defaultdict
from anvil.util.cache import cache
# from urllib.parse import urlparse
from datetime import datetime

from anvil.terra.api import get_entities, get_schema
from anvil.terra.subject import subject_factory
from anvil.terra.sample import sample_factory
import json

def _blobs(bucket_name, user_project, workspace_id):
    """Retrieve all blobs in terra bucket associated with workspace, dict keyed by object url.

    Checks workspace.bucketName and workspace.project_files
    :type project: str or None
    :param project: the project which the client acts on behalf of. Will be
                    passed when creating a topic.  If not passed,
                    falls back to the default inferred from the environment.
    """
    # Instantiates a google client, & get all blobs in bucket
    storage_client = storage.Client(user_project)
    _blobs = cache.get(bucket_name)
    if not _blobs:
        logging.getLogger(__name__).info(f"bucket {bucket_name} not in cache, fetching from google.")
        bucket = storage_client.bucket(bucket_name, user_project=user_project)
        # get subset of data
        _blobs = {}
        try:
            for b in bucket.list_blobs(fields='items(size, etag, crc32c, name, timeCreated),nextPageToken'):
                name = f"gs://{bucket_name}/{b.name}"
                # cache.put(name, {'size': b.size, 'etag': b.etag, 'crc32c': b.crc32c, 'time_created': b.time_created, 'name': name})
                _blobs[name] = AttrDict({'size': b.size, 'etag': b.etag, 'crc32c': b.crc32c, 'time_created': b.time_created, 'name': name})
            cache.put(bucket_name, _blobs)
        except Exception as e:
            logging.getLogger(__name__).error(f"{workspace_id} {bucket_name} {e}")
    return _blobs


class Workspace():
    """Represent terra workspace."""

    def __init__(self, *args, user_project=None, drs_file_path=None, data_ingestion_tracker=None):
        """Pass all args to AttrDict, set id for cacheing."""
        self.attributes = AttrDict(*args)
        assert user_project, "Must have user_project"
        self._user_project = user_project
        self._logger = logging.getLogger(__name__)
        self.id = self.attributes.workspace.name
        self._subjects = None
        self._schemas = None
        self._samples = None
        self._blobs = None
        self._project_files = None
        self._missing_project_files = None
        self.missing_sequence = False
        self.drs_file_path = drs_file_path
        self._already_logged = []
        self.data_ingestion_tracker_path = data_ingestion_tracker
        self._data_ingestion_tracker = None

    @property
    def subjects(self):
        """Return raw subjects from terra."""
        if not self._subjects:
            self._subjects = [subject_factory(s, workspace=self, samples=self.samples) for s in self._get_entities(self.subject_property_name)]
        return self._subjects

    @property
    def samples(self):
        """Return raw samples from terra indexed by subject_id."""
        if not self._samples:
            self._samples = defaultdict(list)
            blobs = _blobs(self.attributes.workspace['bucketName'], self._user_project, self.id)
            logging.getLogger(__name__).debug(f"bucket {self.attributes.workspace['bucketName']} billing {self._user_project} retrieved.")
            sequencing = self._get_entities('sequencing')
            logging.getLogger(__name__).debug(f"retrieved sequencing in {self.id}.")
            # TODO - start optimize me
            for s in self._get_entities('sample'):
                s = sample_factory(s, workspace=self, blobs=blobs, sequencing=sequencing, drs_file_path=self.drs_file_path)
                self._samples[s.subject_id].append(s)
                if s.missing_sequence:
                    self.missing_sequence = s.missing_sequence
            # TODO - end optimize me
            logging.getLogger(__name__).debug(f"created samples in {self.id}.")
        return self._samples

    @property
    def project_files(self):
        """Return files associated with project attributes."""
        if not self._project_files:
            self._project_file_blobs()
        return self._project_files

    @property
    def project_files_attributes(self):
        """Find attributes that are files."""
        _files = {}
        for k, v in self.attributes.workspace.items():
            if isinstance(v, str) and v.startswith('gs://'):
                _files[k] = v
        return _files

    @property
    def missing_project_files(self):
        """Return files associated with project attributes where gs:// blobs are missing."""
        if not self._missing_project_files:
            self._project_file_blobs()
        return self._missing_project_files

    def _project_file_blobs(self):
        """Retrieve all blobs in terra bucket associated with workspace, dict keyed by object url."""
        project_files = _project_files(self.attributes.workspace)
        project_files_keys = project_files.keys()
        if len(project_files_keys) > 0:
            # project_buckets = set([urlparse(f).netloc for f in project_files.values()])
            project_blobs = {}
            # for project_bucket in project_buckets:
            #     project_blobs = {**project_blobs, **_bucket_contents(self._user_project, project_bucket)}
            # project_blobs = AttrDict(project_blobs)
            for k, v in project_files.items():
                b = project_blobs.get(v, None)
                if not b:
                    if not self._missing_project_files:
                        self._missing_project_files = AttrDict({})
                    self._missing_project_files[k] = {'value': v, 'blob': None}
                else:
                    project_files[k] = AttrDict({'value': v, 'blob': b})
            if self._missing_project_files:
                for k in self._missing_project_files:
                    del project_files[k]
            self._project_files = AttrDict(project_files)

    @property
    def missing_samples(self):
        """Test if any missing samples."""
        missing = [s for s in self.subjects if len(s.samples) == 0]
        if len(missing) == 0:
            return None
        return missing

    @property
    def missing_subjects(self):
        """Test if any missing subjects."""
        return len(self.subjects) == 0

    @property
    def missing_blobs(self):
        """Test if any sample missing blobs."""
        missing = [s for s in self.subjects if len([sa for sa in s.samples if sa.missing_blobs]) > 0]
        if len(missing) == 0:
            return None
        return missing

    @property
    def inconsistent_entityName(self):
        """Test if any samples with inconsistent_entityName."""
        a = [s for s in self.subjects if len([sa for sa in s.samples if sa.inconsistent_entityName]) > 0]
        if len(a) == 0:
            return None
        return a

    @property
    def inconsistent_subject(self):
        """Test if any samples with inconsistent_entityName."""
        a = [s for s in self.subjects if len([sa for sa in s.samples if sa.inconsistent_subject]) > 0]
        if len(a) == 0:
            return None
        return a

    @property
    def blob_sizes(self):
        """Aggregate sample blob sizes by property name."""
        _blob_sizes = defaultdict(int)
        for s in self.subjects:
            for sa in s.samples:
                for k, v in sa.blob_sizes.items():
                    _blob_sizes[k] += v
        return _blob_sizes

    @property
    def file_histogram(self):
        """Aggregate sample sizes by property date."""
        def histogram():
            return {'count': 0, 'size': 0, 'date': None}
        _file_histogram = defaultdict(histogram)

        for s in self.subjects:
            for sa in s.samples:
                for blob in sa.blobs.values():
                    time_created = str(blob['time_created'])
                    date_created = datetime.fromisoformat(time_created).date().isoformat()
                    _file_histogram[date_created]['count'] += 1
                    _file_histogram[date_created]['date'] = date_created
                    _file_histogram[date_created]['size'] += blob['size']
        return _file_histogram

    @property
    def files(self):
        """Aggregate sample sizes by type."""
        def f():
            return {'count': 0, 'size': 0, 'type': None}
        _files = defaultdict(f)

        for s in self.subjects:
            for sa in s.samples:
                for blob in sa.blobs.values():
                    # get extension
                    type = blob['name'].replace('.gz', '')
                    type = type.split('/')[-1].split('.')[-1]
                    _files[type]['count'] += 1
                    _files[type]['type'] = type.title()
                    _files[type]['size'] += blob['size']
        return _files

    def _get_entities(self, entity_name):
        """Return all entities in a workspace."""
        return get_entities(self.attributes.workspace.namespace, self.attributes.workspace.name, entity_name)

    @property
    def subject_property_name(self):
        """Return name of subject/participant entity, depending on schema."""
        subject_property_name = 'subject'
        if 'participant' in self.schemas.keys():
            subject_property_name = 'participant'
        return subject_property_name

    @property
    def name(self):
        """Return name of workspace."""
        return self.attributes.workspace.name

    @property
    def schemas(self):
        """Return schema for workspace."""
        if not self._schemas:
            self._schemas = get_schema(self.attributes.workspace.namespace, self.attributes.workspace.name)
        return self._schemas

    @property
    def subject_schema(self):
        """Return schema for workspace subject."""
        return self.schemas.get(self.subject_property_name, None)

    @property
    def sample_schema(self):
        """Return schema for workspace sample."""
        if not hasattr(self, '_already_logged'):
            self._already_logged = []
        if not self._schemas:
            if 'missing-schema' not in self._already_logged:
                logging.error(f"{self.id} - missing schema.")
                self._already_logged.append('missing-schema')
            return None
        if 'sample' not in self._schemas:
            if 'missing-sample-schema' not in self._already_logged:
                logging.error(f"{self.id} - no sample in schema? {self._schemas}")
                self._already_logged.append('missing-sample-schema')
            return None
        return self._schemas['sample']

    def __repr__(self):
        """Return attributes."""
        return str(self.attributes)

    @property
    def data_type(self):
        """Return data type or None."""
        try:
            return self.attributes.workspace.attributes['library:datatype']['items']
        except Exception as e:
            self._logger.debug(f"data_category {e}")
            return None

    @property
    def data_category(self):
        """Return data category or None."""
        try:
            return self.attributes.workspace.attributes['library:dataCategory']['items']
        except Exception as e:
            self._logger.debug(f"data_category {e}")
            return None

    @property
    def problems(self):
        """Flag all problems."""
        return AttrDict({
            'inconsistent_entityName': self.inconsistent_entityName is not None,
            'inconsistent_subject': self.inconsistent_subject is not None,
            'missing_blobs': self.missing_blobs is not None,
            'missing_samples': self.missing_samples is not None or len(self.samples) == 0,
            'missing_project_files': self.missing_project_files is not None,
            'missing_subjects': self.missing_subjects,
            'missing_schema': self.schemas is None or len(self.schemas.keys()) == 0,
            'missing_sequence': self.missing_sequence,
            'missing_diseaseOntology': self.diseaseOntologyId is None,
            'missing_fromSpreadSheet': self.data_ingestion_tracker_attributes is None,
        })

    @property
    def dashboard_view(self):
        """Format for portal view."""
        return AttrDict({
            'file_histogram': [h for h in self.file_histogram.values()],
            'files': [f for f in self.files.values()],
            'nodes': [
                {
                    "type": "Project",
                    "count": 1
                },
                {
                    "type": "Subject",
                    "count": len(self.subjects)
                },
                {
                    "type": "Samples",
                    # samples is a dict keyed by subject id, sum the len of each subject's sample list
                    "count": sum([len(sl) for sl in list(self.samples.values())])
                },
            ],
            'size': sum([f['size']for f in self.files.values()]),
            'disease_ontology_id': self.diseaseOntologyId,
            'project_id': self.name,
            'public': self.attributes['public'],
            'createdDate': self.attributes.workspace.createdDate,
            'lastModified': self.attributes.workspace.lastModified,
            'data_type': self.data_type,
            'data_category': self.data_category,
            'problems': self.problems
        })

    @property
    def investigator(self):
        """Deduce investigator name."""
        _investigator = self.attributes.workspace.attributes.get('study_pi', None)
        if not _investigator:
            _investigator = self.attributes.workspace.attributes.get("library:datasetOwner", None)
            if _investigator == 'NA':
                _investigator = self.attributes.workspace.attributes.get('library:datasetCustodian', None)
        return _investigator

    @property
    def accession(self):
        """Deduce accession."""
        return self.attributes.workspace.attributes.get("library:datasetVersion", None)

    @property
    def institute(self):
        """Deduce institute."""
        _institute = self.attributes.workspace.attributes.get("library:institute", None)
        if _institute and 'items' in _institute:
            return _institute['items'][0]
        return _institute

    @property
    def data_ingestion_tracker_attributes(self):
        if self._data_ingestion_tracker:
            return self._data_ingestion_tracker
        rows = json.load(open(self.data_ingestion_tracker_path))
        for row in rows:
            if row['workspace_name'] == self.name:
                self._data_ingestion_tracker = row
        if not self._data_ingestion_tracker and 'missing_fromSpreadSheet' not in self._already_logged:        
            logging.getLogger(__name__).error(f"data_ingestion_tracker_attributes missing from spreadsheet {self.name} (supressing error for this workspace)")
            self._already_logged.append('missing_fromSpreadSheet')
        return self._data_ingestion_tracker

    @property
    def diseaseOntologyId(self):
        """Deduce disease."""
        _diseaseOntologyID = self.attributes.workspace.attributes.get('diseaseOntologyID', None)
        if not _diseaseOntologyID:
            _diseaseOntologyID = self.attributes.workspace.attributes.get('library:diseaseOntologyID', None)
        if not _diseaseOntologyID and self.data_ingestion_tracker_attributes:
            _diseaseOntologyID = self.data_ingestion_tracker_attributes['diseaseOntologyId']
        if _diseaseOntologyID:
            _diseaseOntologyID = _diseaseOntologyID.split('/')[-1].replace('_', ':')            
        else:
            logging.debug(f"{self.id} missing diseaseOntologyID")
        return _diseaseOntologyID


def _project_files(w):
    """Deduce attributes that are files."""
    return {k: v for k, v in w['attributes'].items() if isinstance(v, str) and v.startswith('gs://')}


# @memoize
# def _bucket_contents(user_project, bucket_name):
#     """Fetch bucket contents."""
#     storage_client = storage.Client(project=user_project)
#     project_blobs = {}
#     project_bucket = storage_client.bucket(bucket_name, user_project=user_project)
#     for b in list(project_bucket.list_blobs()):
#         name = f"gs://{project_bucket.name}/{b.name}"
#         project_blobs[name] = {'size': b.size, 'etag': b.etag, 'crc32c': b.crc32c, 'time_created': b.time_created, 'name': name}
#     return project_blobs


class CMGWorkspace(Workspace):
    """Extend Workspace class."""

    def __init__(self, *args, **kwargs):
        """Call super."""
        super().__init__(*args, **kwargs)

    @property
    def accession(self):
        """Deduce accession."""
        _accession = super().accession
        if not _accession:
            _accession = self.attributes.workspace.attributes.get("study_accession", None)
        return _accession


class CCDGWorkspace(Workspace):
    """Extend Workspace class."""

    def __init__(self, *args, **kwargs):
        """Call super."""
        super().__init__(*args, **kwargs)


def workspace_factory(*args, **kwargs):
    """Return a specialized Workspace class instance."""
    name = args[0]['workspace']['name']
    if 'CCDG' in name.upper():
        return CCDGWorkspace(*args, **kwargs)
    if 'CMG' in name.upper():
        return CMGWorkspace(*args, **kwargs)
    return Workspace(*args, **kwargs)
    # if 'GTEX' in kwargs['workspace'].name.upper():
    #     return GTExSubject(*args, **kwargs)
    # if '1000G-HIGH-COVERAGE' in kwargs['workspace'].name.upper():
    #     return ThousandGenomesSubject(*args, **kwargs)
    # if 'ANVIL_EMERGE' in kwargs['workspace'].name.upper():
    #     return eMERGESUbject(*args, **kwargs)
    raise Exception('Not implemented')



import os
import click
from anvil.etl.transformers.normalizer import ontologies, _recursive_default_dict, fetch_workspace_names, \
    write_workspace

from anvil.etl.transformers.normalizer import normalize, analyze
from anvil.etl.transformers.fhir_writer import write
import json
from multiprocessing import Pool

import logging 
logger = logging.getLogger(__name__)


@click.group(name='transform')
@click.pass_context
def transform(ctx):
    """Transform data commands."""
    # Set log level in our & dependencies loggers
    logger.setLevel(ctx.obj['log_level'])


@transform.command(name='normalize')
@click.option('--consortium', default=None, help='Filter, only this consortium.')
@click.option('--workspace', default=None, help='Filter, only this workspace.')
@click.option('--validate_buckets/--no-validate_buckets', default=True, help='Check gs:// urls against bucket contents.', show_default=True)
@click.pass_context
def _normalize(ctx, consortium, workspace, validate_buckets):
    """Normalize workspace and write summary to stdout."""
    if not consortium:
        consortium = _consortium_from_workspace(ctx.obj['config'], workspace)
        workspace_names = [(consortium, workspace)]
    else:
        workspace_names = fetch_workspace_names(ctx.obj['output_path'], requested_consortium_name=consortium, workspace_name=workspace)

    with Pool(maxtasksperchild=1) as pool:
        for consortium_name, workspace_name in workspace_names:
            pool.starmap(write_workspace, [(ctx.obj['output_path'], consortium, workspace_name, validate_buckets, ctx.obj['config'])])


def _print_analysis(output_path, consortium, workspace, details, validate_buckets, config, file_name):
    """Process worker."""
    analysis = analyze(output_path, consortium, workspace, details, validate_buckets, config)
    if analysis == {}:
        logger.error(('no.analysis', workspace))
        return
    _json = json.dumps(analysis[consortium][workspace], separators=(',', ':'))
    with open(file_name, "a") as output_stream:
        output_stream.write(_json)    
        output_stream.write('\n')    
    logger.info(f"wrote {workspace}")


def _consortium_from_workspace(config, workspace_name):    
    """Extract consortium from config."""
    if not workspace_name:
        return None
    for k in config['consortiums']:
        if k.lower() in workspace_name.lower():
            return k
    assert False, f"no consortium for {workspace_name}"


@transform.command(name='analyze')
@click.option('--consortium', default=None, help='Filter, only this consortium.')
@click.option('--workspace', default=None, help='Filter, only this workspace.')
@click.option('--details', default=False, help='Include error details.', show_default=True, is_flag=True)
@click.option('--validate_buckets/--no-validate_buckets', default=True, help='Check gs:// urls against bucket contents.', show_default=True)
@click.pass_context
def _analyze(ctx, consortium, workspace, details, validate_buckets):
    """Retrieve workspace and write summary <data>/analysis.ndjson."""
    if workspace:
        if not consortium:
            consortium = _consortium_from_workspace(ctx.obj['config'], workspace)
        workspace_names = [(consortium, workspace)]
    else:
        workspace_names = fetch_workspace_names(ctx.obj['output_path'], requested_consortium_name=consortium, workspace_name=workspace)

    import os
    import contextlib
    with contextlib.suppress(FileNotFoundError):
        os.remove(f"{ctx.obj['output_path']}/analysis.ndjson")

    file_name = f"{ctx.obj['output_path']}/analysis.ndjson"
    logging.info(f"writing to {file_name}")
    with Pool(maxtasksperchild=1) as pool:
        for consortium_name, workspace_name in workspace_names:
            pool.starmap(_print_analysis, [(ctx.obj['output_path'], consortium_name, workspace_name, details, validate_buckets, ctx.obj['config'], file_name)])


def _fhir_transform(workspace_name, output_path, config, validate_buckets, details):
    for consortium_name, _workspace in normalize(
            output_path, workspace_name=workspace_name,
            requested_consortium_name=None,
            validate_buckets=validate_buckets, config=config):
        write(consortium_name, _workspace, output_path, details, config)


@transform.command(name='fhir')
@click.option('--consortium', default=None, help='Filter, only this consortium.')
@click.option('--workspace', default=None, help='Filter, only this workspace')
@click.option('--validate_buckets/--no-validate_buckets', default=True, help='Check gs:// urls against bucket contents.', show_default=True)
@click.option('--details', default=False, help='Include error details.', show_default=True, is_flag=True)
@click.pass_context
def _fhir(ctx, consortium, workspace, validate_buckets, details):
    """Normalize workspace and write as FHIR to file system (takes several minutes)."""
    if workspace:
        workspace_names = [workspace]
    else:
        workspace_names = fetch_workspace_names(ctx.obj['output_path'], requested_consortium_name=consortium, workspace_name=workspace)

    # with Pool(maxtasksperchild=1) as pool:
    #     for consortium_name, workspace_name in workspace_names:
    #         pool.starmap(_fhir_transform, [(workspace_name, ctx.obj['output_path'], ctx.obj['config'], validate_buckets, details,)])

    for consortium_name, workspace_name in workspace_names:
        _fhir_transform(workspace_name, ctx.obj['output_path'], ctx.obj['config'], validate_buckets, details,)



"""Cache items into sqlite."""
import sqlite3
import json
import logging
from datetime import date, datetime
from collections import defaultdict

logger = logging.getLogger(__name__)


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code."""
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


class Entities():
    """Cache items in sqlite."""

    def __init__(self, path):
        """Set up sqlite db."""
        # works better w/ flask
        self._path = path
        self._conn = sqlite3.connect(path, check_same_thread=True)
        self._conn.row_factory = dict_factory
        cur = self._conn.cursor()
        cur.executescript("""
        CREATE TABLE IF NOT EXISTS vertices (
            key text PRIMARY KEY,
            label text NOT NULL,
            json text NOT NULL
        );
        CREATE TABLE IF NOT EXISTS edges (
            src text,
            dst text,
            src_name text,
            dst_name text
        );
        """)
        self._conn.commit()

        # optimize for single thread speed
        self._conn.execute('PRAGMA synchronous = OFF')
        self._conn.execute('PRAGMA journal_mode = OFF')
        # double cache size
        self._conn.execute('PRAGMA cache_size = -6000')
        self._conn.commit()
        self._conn.close()
        self._conn = sqlite3.connect(path, check_same_thread=False, isolation_level='DEFERRED')
        self._conn.row_factory = dict_factory
        self._cursor = None
        self._put_count = 0

    def get(self, key):
        """Retrieve an item."""

        logger.debug(f"? {key}")
        cur = self._conn.cursor()
        data = cur.execute("SELECT json FROM vertices where key=? ", (key,)).fetchone()
        if data:
            data = json.loads(data['json'])
            logger.debug(f"hit {key}")
        else:
            logger.debug(f"miss {key}")
        cur.close()
        return data

    @property
    def cursor(self):
        if not self._cursor:
            self._cursor = self._conn.cursor()
        return self._cursor

    def get_by_label(self, label):
        """Retrieve all items."""

        logger.debug(f"? {label}")
        rows = self.cursor.execute("SELECT json FROM vertices where label=? order  by key ", (label,)).fetchall()
        for row in rows:
            yield json.loads(row['json'])

    def commit(self, force=False):
        """Commit if put count over limit."""
        if self._put_count > 1000 or force:
            self._conn.commit()
            self._cursor = None
            self._put_count = 0

    def put(self, key, label, data):
        """Save an item."""
        logger.debug(f"put {key}")
        self.cursor.execute("REPLACE into vertices values (?, ?, ?);", (key, label, json.dumps(data, default=json_serial)))
        self._put_count += 1
        self.commit()

    def put_edge(self, src, dst, src_name, dst_name):
        """Save edge."""
        logger.debug(f"put edge {src_name}/{src} -> {dst_name}/{dst}")
        self.cursor.execute("REPLACE into edges values (?, ?, ?, ?);", (src, dst, src_name, dst_name))
        self._put_count += 1
        self.commit()

    def get_edges(self, src, src_name):
        """Retrieves all edges."""
        destination_edges = self.cursor.execute("SELECT dst, dst_name FROM edges where src= ? and src_name= ? ", (src, src_name, )).fetchall()
        destination_vertices = defaultdict(list)
        for de in destination_edges:
            destination_vertices[de['dst_name']].append(self.get(de['dst']))
        return destination_vertices

    def get_edges_by_label(self, src, src_name, dst_name, limit=None):
        """Retrieves all edges for a particular label."""
        if limit:
            limit = f"LIMIT {limit}"
        else:
            limit = ''
        destination_edges = self.cursor.execute(f"SELECT dst, dst_name FROM edges where src = ? and src_name = ? and dst_name = ? {limit}", (src, src_name, dst_name, )).fetchall()
        destination_vertices = {de['dst_name']: self.get(de['dst']) for de in destination_edges}
        return destination_vertices

    def index(self):
        logger.info('Indexing')
        cur = self._conn.cursor()
        cur.executescript("""
        CREATE  INDEX IF NOT EXISTS items_key ON vertices(key);
        CREATE  INDEX IF NOT EXISTS label_key ON vertices(label);
        CREATE UNIQUE INDEX IF NOT EXISTS edges_src_dst ON edges(src, dst, src_name, dst_name);
        CREATE  INDEX IF NOT EXISTS edges_dst ON edges(dst);
        """)
        self._conn.commit()



import subprocess
from subprocess import Popen, PIPE
import os
import logging
logger = logging.getLogger(__name__)


def run_cmd(command_line):
    """Run a command line, return stdout."""
    try:
        return subprocess.check_output(command_line, shell=True).decode("utf-8").rstrip()
    except Exception as e:
        logger.error(e)




import json
import os

import requests

from anvil.etl.utilities.shell_helper import run_cmd

import click
import logging

from anvil.etl.utilities.shell_helper import run_cmd

logger = logging.getLogger(__name__)


@click.group(name='ig')
@click.pass_context
def ig_cli(ctx):
    pass  # dummy


@ig_cli.command(name='create')
@click.pass_context
def create(ctx):
    """Download and setup IG on local file system, copy to bucket."""

    download_script = """
    mkdir -p  $OUTPUT_PATH/fhir/IG
    # clean up old
    rm -f $OUTPUT_PATH/fhir/IG/*.json || true 2> /dev/null
    rm -f $OUTPUT_PATH/fhir/IG/*.internals || true 2> /dev/null
    rm -f $OUTPUT_PATH/definitions.json.zip || true 2> /dev/null
    rm -f $OUTPUT_PATH/expansions.json.zip || true 2> /dev/null
    # copy from IG build
    curl https://nih-ncpi.github.io/ncpi-fhir-ig/definitions.json.zip --output $OUTPUT_PATH/definitions.json.zip 2> /dev/null 
    curl https://nih-ncpi.github.io/ncpi-fhir-ig/expansions.json.zip --output $OUTPUT_PATH/expansions.json.zip 2> /dev/null

    unzip $OUTPUT_PATH/definitions.json.zip -d $OUTPUT_PATH/fhir/IG
    unzip $OUTPUT_PATH/expansions.json.zip  -d $OUTPUT_PATH/fhir/IG

    # delete extraneous
    rm $OUTPUT_PATH/fhir/IG/*.internals

    """
    # https://cloud.google.com/healthcare/docs/how-tos/fhir-profiles#configure_your_implementation_guide
    run_cmd(download_script)

    # adjust for google
    output_path = ctx.obj["output_path"]
    ig_path = f'{output_path}/fhir/IG/ImplementationGuide-NCPI-FHIR-Implementation-Guide.json'
    ig = json.load(open(ig_path, 'r'))
    # items to add to global
    structure_definitions = [r['reference']['reference'] for r in ig['definition']['resource'] if
                             'StructureDefinition' in r['reference']['reference']]

    ig_global = []
    for _id in structure_definitions:
        _id = _id.replace('/', '-')
        sd_path = f'{output_path}/fhir/IG/{_id}.json'
        sd = json.load(open(sd_path, 'r'))
        if sd['kind'] != 'resource':
            continue
        ig_global.append({'type': sd['type'], 'profile': sd['url']})

    ig['global'] = ig_global
    # logger.info(f"added to 'global' {[g['type'] for g in ig['global']]}")
    json.dump(ig, open(ig_path, 'w'), separators=(',', ':'))

    move_ig_to_bucket = """
    gsutil -m cp -J -r $OUTPUT_PATH/fhir/IG    gs://$GOOGLE_BUCKET/fhir
    # also need to include all dependencies
    curl -s http://hl7.org/fhir/us/core/STU3.1.1/ImplementationGuide-hl7.fhir.us.core.json | gsutil cp - gs://$GOOGLE_BUCKET/IG/ImplementationGuide-hl7.fhir.us.core.json

    curl -s https://www.hl7.org/fhir/definitions.json.zip -o /tmp/definitions.json.zip
    unzip -p /tmp/definitions.json.zip valuesets.json > /tmp/valuesets.json
    cat /tmp/valuesets.json | gsutil cp - gs://$GOOGLE_BUCKET/fhir/IG/valuesets/valuesets.json
    rm /tmp/definitions.json.zip
    rm /tmp/valuesets.json        
    """
    run_cmd(move_ig_to_bucket)

    logger.debug(run_cmd("gsutil ls gs://$GOOGLE_BUCKET/fhir/IG"))

    logger.info(f"IG setup complete and copied to gs://{os.environ['GOOGLE_BUCKET']}/fhir/IG.")


@ig_cli.command(name='delete')
@click.pass_context
def delete(ctx):
    """Remove IG from local file system and bucket."""
    delete_script = """
    rm -r  $OUTPUT_PATH/fhir/IG  || true > /dev/null
    gsutil -m rm -r  gs://$GOOGLE_BUCKET/fhir/IG 
    """
    run_cmd(delete_script)


def _enable_ig(ctx, data_store, token):
    """Enable IG in datastores"""
    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json; charset=utf-8"
    }
    project = os.environ['GOOGLE_PROJECT']
    location = os.environ['GOOGLE_LOCATION']
    data_set = os.environ['GOOGLE_DATASET']
    url = f"https://healthcare.googleapis.com/v1beta1/projects/{project}/locations/{location}/datasets/{data_set}/fhirStores/{data_store}?updateMask=validationConfig"
    response = requests.patch(
        url=url,
        headers=headers,
        json={
            "validationConfig": {
                "enabledImplementationGuides": ["https://ncpi-fhir.github.io/ncpi-fhir-ig/ImplementationGuide/NCPI-FHIR-Implementation-Guide"]
            }
        }
    )
    if response.status_code == 200:
        logger.info(f"IG enabled on {data_store}")
    else:
        logger.info(("could.not.enable.ig", data_store, response.status_code, response.text))


def _get_ig(ctx, data_store, token):
    """Get IG info"""
    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json; charset=utf-8"
    }
    project = os.environ['GOOGLE_PROJECT']
    location = os.environ['GOOGLE_LOCATION']
    data_set = os.environ['GOOGLE_DATASET']
    url = f"https://healthcare.googleapis.com/v1beta1/projects/{project}/locations/{location}/datasets/{data_set}/fhirStores/{data_store}"
    response = requests.get(
        url=url,
        headers=headers,
    )
    response.raise_for_status()
    return response.json()



"""Instances of this class handle authorizing and talking to Google Healthcare API FHIR Service."""


import logging
import threading
from urllib.parse import urlparse
from fhirclient import client
from fhirclient.models.meta import Meta
from fhirclient.models.bundle import Bundle

from anvil.fhir.smart_auth import GoogleFHIRAuth

logger = logging.getLogger(__name__)


class FHIRClient(client.FHIRClient):
    """Instances of this class handle authorizing and talking to Google Healthcare API FHIR Service.

    Parameters:
        See https://github.com/smart-on-fhir/client-py/blob/master/fhirclient/client.py#L19

    Returns:
        Instance of client, with injected authorization method

    Examples: ::
        from anvil.fhir.client import FHIRClient
        settings = {
            'app_id': 'my_web_app',
            'api_base': 'https://healthcare.googleapis.com/v1/projects/gcp-testing-308520/locations/us-east4/datasets/testset/fhirStores/fhirstore/fhir'
        }
        smart = FHIRClient(settings=settings)
        assert smart.ready, "server should be ready"
        # search for all ResearchStudy
        import fhirclient.models.researchstudy as rs
        [s.title for s in rs.ResearchStudy.where(struct={}).perform_resources(smart.server)]
        >>>
        ['1000g-high-coverage-2019', 'my NCPI research study example']


    """

    def __init__(self, *args, **kwargs):
        """Pass args to super, adds GoogleFHIRAuth authenticator, prepares connection."""
        super(FHIRClient, self).__init__(*args, **kwargs)
        client_major_version = int(client.__version__.split('.')[0])
        assert client_major_version >= 4, f"requires version >= 4.0.0 current version {client.__version__} `pip install -e git+https://github.com/smart-on-fhir/client-py#egg=fhirclient`"
        self.server.auth = GoogleFHIRAuth()
        self.server.session.hooks['response'].append(self.server.auth.handle_401)
        self.prepare()
        assert self.ready, "server should be ready"


class DispatchingFHIRClient(client.FHIRClient):
    """Instances of this class handle authorizing and talking to Google Healthcare API FHIR Service.

    Parameters:
        See https://github.com/smart-on-fhir/client-py/blob/master/fhirclient/client.py#L19

    :param settings.api_bases: The servers against which to perform the search **settings.api_base ignored**
    :param access_token: Optional access token, if none provided `gcloud auth print-access-token` is used

    Returns:
        Instance of client, with injected authorization method

    Examples: ::
        from anvil.fhir.client import DispatchingFHIRClient
        from fhirclient.models.researchstudy import ResearchStudy
        from collections import defaultdict
        from pprint import pprint

        settings = {
            'app_id': 'my_web_app',
            'api_bases': [
                'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-11-329119/locations/us-west2/datasets/anvil-test/fhirStores/public/fhir',
                'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-11-329119/locations/us-west2/datasets/anvil-test/fhirStores/pending/fhir',
            ]
        }
        smart = DispatchingFHIRClient(settings=settings)

        # search for all ResearchStudy, index by source
        studies = defaultdict(list)
        for s in ResearchStudy.where(struct={'_count':'1000'}).perform_resources(smart.server):
            studies[s.meta.source].append(s)

        pprint({k: len(v) for k,v in studies.items()})
        >>>
        {'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-11-329119/locations/us-west2/datasets/anvil-test/fhirStores/pending/fhir/': 259,
        'https://healthcare.googleapis.com/v1beta1/projects/fhir-test-11-329119/locations/us-west2/datasets/anvil-test/fhirStores/public/fhir/': 393}

    """

    def __init__(self, *args, **kwargs):
        """Pass args to super, patches `perform` to our dispatching version."""
        # use the first entry as 'our' server
        _settings = dict(kwargs['settings'])
        api_base = _settings['api_bases'].pop()
        _settings['api_base'] = api_base
        kwargs['settings'] = _settings

        # grab a token if passed
        access_token = None
        if 'access_token' in kwargs:
            access_token = kwargs['access_token']
            del kwargs['access_token']

        # normal setup with our authenticator
        super(DispatchingFHIRClient, self).__init__(*args, **kwargs)
        client_major_version = int(client.__version__.split('.')[0])
        assert client_major_version >= 4, f"requires version >= 4.0.0 current version {client.__version__} `pip install -e git+https://github.com/smart-on-fhir/client-py#egg=fhirclient`"
        self.server.auth = GoogleFHIRAuth(access_token=access_token)
        self.server.session.hooks['response'].append(self.server.auth.handle_401)
        self.prepare()
        assert self.ready, "server should be ready"

        # set up an array of FHIRClients, including this instance, in self._clients
        # re-use authenticator
        self._clients = [self]
        self._api_bases = _settings['api_bases']
        for api_base in self._api_bases:
            __settings = dict(_settings)
            __settings['api_base'] = api_base
            _client = client.FHIRClient(settings=__settings)
            _client.server.auth = self.server.auth
            _client.server.session.hooks['response'].append(self.server.auth.handle_401)
            _client.prepare()
            self._clients.append(_client)

        # monkey patch search perform if we haven't already
        from fhirclient.models.fhirsearch import FHIRSearch
        if not hasattr(FHIRSearch, '_anvil_patch'):
            FHIRSearch._anvil_patch = True
            logger.debug("******** Needs patching ********")
            original_perform = FHIRSearch.perform
            me = self

            def _perform(self, server):
                """Dispatch query to api_bases."""

                def _worker(self, server, _results):
                    """Dispatches request to underlying class, return an entry indexed by base uri.

                    Sets bundle.meta.source
                    See https://www.hl7.org/fhir/resource-definitions.html#Meta.source

                    :param server: The server against which to perform the search
                    :_results: Result of operation added to this array
                    """
                    logger.debug(f"worker starting {server.base_uri}")
                    result = original_perform(self, server)
                    logger.debug(f"worker got {result}")
                    while result:

                        # add source to meta if it doesn't already exist
                        if not result.meta:
                            result.meta = Meta()
                        if not result.meta.source:
                            result.meta.source = server.base_uri
                        _results.append(result)

                        # follow `next` link for pagination
                        if hasattr(result, 'link'):
                            _next = next((lnk.as_json() for lnk in result.link if lnk.relation == 'next'), None)
                            result = None
                            if _next:
                                logger.debug(f"has next {_next}")
                                # request_json takes a full path & query (not host)
                                parts = urlparse(_next['url'])
                                assert len(parts.query) > 0, parts
                                path = f"{parts.path}?{parts.query}"
                                logger.debug(f"attempting next {path}")
                                res = server.request_json(path)
                                result = Bundle(res)
                                result.origin_server = server
                        else:
                            result = None
                    logger.debug(f"worker done {result}")

                logger.debug("starting threads")
                workers = []
                results = []
                for _client in me._clients:
                    workers.append(
                        threading.Thread(target=_worker, args=(self, _client.server, results, ))
                    )
                # Start workers.
                for w in workers:
                    w.start()

                # Wait for workers to quit.
                logger.debug("waiting for results.")
                for w in workers:
                    w.join()
                logger.debug(f"all workers done. {len(results)}")
                return results
            # monkey patch
            FHIRSearch.perform = _perform

            # since perform returns an array, patch _perform_resources as well.
            def _perform_resources(self, server):
                """Perform the search by calling `perform`, then extracts all Bundle entries and returns a list of Resource instances.

                Sets resource.meta.source
                See https://www.hl7.org/fhir/resource-definitions.html#Meta.source

                :param server: The server against which to perform the search
                :returns: A list of Resource instances
                """
                # flatten into an array of resources
                bundles = self.perform(server)
                resources = []
                if bundles is not None:
                    if not isinstance(bundles, list):
                        bundles = [bundles]
                    for bundle in bundles:
                        if bundle.entry:
                            for entry in bundle.entry:
                                if not entry.resource.meta:
                                    entry.resource.meta = Meta()
                                if not entry.resource.meta.source:
                                    entry.resource.meta.source = bundle.meta.source
                                resources.append(entry.resource)
                logger.debug("_perform_resources done.")
                return resources
            FHIRSearch.perform_resources = _perform_resources



"""Query dbGap for study information."""

import logging
import requests
import xmltodict
import json
import os
from anvil.util.cache import memoize

logger = logging.getLogger('anvil.dbgap.api')

url_base = 'https://www.ncbi.nlm.nih.gov/projects/gap/cgi-bin/study.cgi?study_id='

# sync with setup.py utility data_ingestion_tracker
DEFAULT_OUTPUT_PATH = f"{os.path.dirname(__file__)}/../data/data_ingestion_tracker.json"


@memoize
def get_study(accession):
    """Return tuple (qualified_accession, schema)."""
    qualified_accession = None
    try:
        r = requests.get(f"{url_base}{accession}", allow_redirects=False)
        assert r.status_code == 302, r.status_code
        qualified_accession = r.headers['location'].split('=')[1]
        assert len(qualified_accession) > 0, f"No qualified study for {accession}"
        r = requests.get(f"https://www.ncbi.nlm.nih.gov/projects/gap/cgi-bin/GetSampleStatus.cgi?study_id={qualified_accession}&rettype=xml")
        assert r.status_code == 200, f"status_code: {r.status_code} text:{r.text}"
        d = xmltodict.parse(r.text)
        return ((qualified_accession, d))
    except Exception as e:
        logger.warn(f"{accession}/{qualified_accession} error: {e}")
        return None


@memoize
def get_projects():
    """Open embedded extract from google spreadsheet, indexed by workspace."""
    with open(DEFAULT_OUTPUT_PATH, 'r') as ins:
        return {s['workspace_name']: s for s in json.load(ins)}


def get_accession(namespace, project_id):
    """Return simple_accession (dbGap phsid)."""
    projects = get_projects()
    key = f"{namespace}/{project_id}"
    if key not in projects:
        return None
    simple_accession = projects[key]['simple_accession']
    if not projects[key]['simple_accession']:
        logger.warning(f"{project_id} missing dbGap accession")
        return None
    if not projects[key]['simple_accession'].startswith('phs'):
        logger.warning(f"{project_id} invalid dbGap {simple_accession}")
        return None
    return simple_accession



import json
import os
from collections import defaultdict
from anvil.etl.utilities.shell_helper import run_cmd


def _extract_workspace_mapping(output_path):
    """Read ResearchStudy identifiers to retrieve the consortium, data_store and id."""
    list_script = f"""
        find {output_path} -name ResearchStudy.ndjson -print
    """
    mapping = {'consortium': defaultdict(list), 'data_store': defaultdict(list), 'workspace_name': defaultdict(list)}
    for path in [line for line in run_cmd(list_script).split('\n')]:
        research_studies = []
        with open(path) as input_stream:
            research_studies.append(json.load(input_stream))
        for research_study in research_studies:
            identifiers = research_study['identifier']
            consortium = next(iter([identifier['value'] for identifier in identifiers if identifier['system'] == "https://anvil.terra.bio/#consortium"]), None)
            workspace = next(iter([identifier['value'] for identifier in identifiers if identifier['system'] == "https://anvil.terra.bio/#workspaces/anvil-datastorage/"]), None)
            data_store = next(iter([identifier['value'] for identifier in identifiers if identifier['system'] == "https://anvil.terra.bio/#FHIR/data-store"]), None)
            _id = research_study['id']
            obj = {'name': workspace, 'id': _id, 'consortium': consortium, 'data_store': data_store}
            mapping['consortium'][consortium].append(obj)
            mapping['data_store'][data_store].append(obj)
            mapping['workspace_name'][workspace].append(obj)
    return mapping


def _log_console_link(logger):
    """Log a message with a link to the GCP console"""
    project = os.environ['GOOGLE_PROJECT']
    location = os.environ['GOOGLE_LOCATION']
    data_set = os.environ['GOOGLE_DATASET']
    logger.info(
        f"Data loading, check GCP console for progress. https://console.cloud.google.com/healthcare/browser/locations/{location}/datasets/{data_set}/operations?project={project}")



"""Represent fhir entity."""

from os import stat
from anvil.transformers.fhir import make_workspace_id, make_identifier
import logging


class Organization:
    """Create fhir entity."""

    class_name = "organization"
    resource_type = "Organization"

    @staticmethod
    def slug(resource):
        """Make id."""
        return make_workspace_id(resource)

    @staticmethod
    def build_entity(workspace):
        """Create fhir entity."""
        study_id = workspace.id
        id = Organization.slug(workspace)
        institute = workspace.institute
        if not institute:
            logging.getLogger(__name__).warning(f'workspace {workspace.id} missing institute')

        parent = f'Organization/{workspace.attributes.reconciler_name.lower()}'

        entity = {
            "resourceType": Organization.resource_type,
            "id": f"{id}",
            "meta": {
                "profile": [
                    "http://hl7.org/fhir/StructureDefinition/Organization"
                ]
            },
            "identifier": [
                {
                    "system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/",
                    "value": study_id,
                },
                {
                    "system": "urn:ncpi:unique-string",
                    "value": f"{id}",
                },
                {
                    "system": "anvil:consortium",
                    "value": f"{workspace.attributes.reconciler_name.lower()}",
                }
            ],
            "name": study_id,
        }

        if institute:
            entity['identifier'].append(
                {
                    "system": "anvil:institute",
                    "value": f"{workspace.institute.lower()}",
                }
            )

        entity['partOf'] = {
            "reference": parent
        }

        return entity


    @staticmethod
    def build_practitioner_org(workspace):
        """Create fhir entity."""
        institute = workspace.institute or 'Unknown'
        if not institute:
            logging.getLogger(__name__).warning(f'workspace {workspace.id} missing institute')
        id = make_identifier(institute)
        entity = {
            "resourceType": Organization.resource_type,
            "id": f"{id}",
            "meta": {
                "profile": [
                    "http://hl7.org/fhir/StructureDefinition/Organization"
                ]
            },
            "identifier": [
                {
                    "system": f"https://anvil.terra.bio",
                    "value": id,
                }
            ],
            "name": institute,
            'partOf': {
                "reference": 'Organization/anvil'
            }
        }
        return entity

    @staticmethod
    def build_consortium_org(workspace):
        """Create fhir entity."""
        id = workspace.attributes.reconciler_name.lower()
        entity = {
            "resourceType": Organization.resource_type,
            "id": f"{id}",
            "meta": {
                "profile": [
                    "http://hl7.org/fhir/StructureDefinition/Organization"
                ]
            },
            "identifier": [
                {
                    "system": f"https://anvil.terra.bio",
                    "value": id,
                }
            ],
            "name": workspace.attributes.reconciler_name,
            'partOf': {
                "reference": 'Organization/anvil'
            }            
        }
        return entity

    @staticmethod
    def build_anvil_org():
        """Create fhir entity."""
        id = 'anvil'
       
        entity = {
            "resourceType": Organization.resource_type,
            "id": f"{id}",
            "meta": {
                "profile": [
                    "http://hl7.org/fhir/StructureDefinition/Organization"
                ]
            },
            "identifier": [
                {
                    "system": f"https://anvil.terra.bio",
                    "value": id,
                }
            ],
            "name": 'AnVIL',
        }
        return entity



"""Transform terra workspaces to FHIR."""
# from anvil.transformers.fhir.attachment import Attachment
from anvil.terra.blob import Blob
from anvil.transformers.fhir.document_reference import DocumentReference
from anvil.transformers.fhir.observation import DiseaseObservation
from anvil.transformers.fhir.organization import Organization
from anvil.transformers.fhir.patient import Patient
from anvil.transformers.fhir.practitioner import Practitioner
from anvil.transformers.fhir.research_study import ResearchStudy
from anvil.transformers.fhir.research_subject import ResearchSubject
from anvil.transformers.fhir.research_study_observation import ResearchStudyObservation
from anvil.transformers.fhir.specimen import Specimen
from anvil.transformers.fhir.task import SpecimenTask
from anvil.transformers.transformer import Transformer
import types


class FhirTransformer(Transformer):
    """Represent terra entities in Fhir.  Transform with .entity() method."""

    def __init__(self, *args, **kwargs):
        """Transform entities."""
        super(FhirTransformer, self).__init__(*args, **kwargs)

    # overrides
    def transform_workspace(self, workspace):
        """Transform workspace."""
        def entity(self):
            practitioner = Practitioner.build_entity(self)
            if practitioner:
                yield practitioner
            practitioner_role = Practitioner.build_practitioner_role(self)
            if practitioner_role:
                yield practitioner_role            

            organization = Organization.build_anvil_org()
            if organization:
                yield organization                
            organization = Organization.build_practitioner_org(self)
            if organization:
                yield organization

            organization = Organization.build_entity(self)
            if organization:
                yield organization
                
            organization = Organization.build_consortium_org(self)
            if organization:
                yield organization


            observation = ResearchStudyObservation.build_entity(self)
            if observation:
                yield observation
            yield ResearchStudy.build_entity(self)
        workspace.entity = types.MethodType(entity, workspace)
        yield workspace

    def transform_subject(self, subject):
        """Transform subject."""
        def entity(self):
            yield Patient.build_entity(self)
            yield ResearchSubject.build_entity(self)
            if self.diseases:
                for d in self.diseases:
                    if d:
                        yield DiseaseObservation.build_entity(self, disease=d)

        subject.entity = types.MethodType(entity, subject)
        yield subject

    def transform_sample(self, sample, subject):
        """Transform sample."""
        _me = self

        def entity(self):
            s = Specimen.build_entity(self, subject)
            yield s
            outputs = []
            for blob in self.blobs.values():
                for b in _me.transform_blob(Blob(blob, sample)):
                    b = DocumentReference.build_entity(b, subject)
                    outputs.append(b)
            # task will inject context into DocumentReference
            yield SpecimenTask.build_entity(inputs=[sample], outputs=outputs, subject=subject)
            for b in outputs:
                yield b
        sample.entity = types.MethodType(entity, sample)
        yield sample

    def transform_blob(self, blob):
        """Transform blob (noop)."""
        def entity(self):
            yield DocumentReference.build_entity(self)
        blob.entity = types.MethodType(entity, blob)
        yield blob



import os

import click
import logging
import yaml
import json

from anvil.etl.extractors.gen3 import DRSReader
from anvil.etl.transform import _consortium_from_workspace
from anvil.etl.transformers import _recursive_default_dict
from anvil.etl.transformers.normalizer import ontologies, normalize, fetch_workspace_names

logger = logging.getLogger(__name__)


@click.group(name='utility')
@click.pass_context
def utility(ctx):
    """Utility commands."""
    # Set log level in our & dependencies loggers
    logger.setLevel(ctx.obj['log_level'])


@utility.command(name='config')
@click.option('--format', type=click.Choice(['json', 'yaml'],case_sensitive=False), default='json', show_default=True)
@click.pass_context
def _config(ctx, format):
    """Print config to stdout."""
    if format == 'yaml':
        print(yaml.dump(ctx.obj['config']))
    else:
        print(json.dumps(ctx.obj['config']))


@utility.command(name='ontologies')
@click.option('--consortium', default=None, help='Filter, only this consortium.')
@click.option('--workspace', default=None, help='Filter, only this workspace.')
@click.option('--details', default=False, help='Include error details.', show_default=True, is_flag=True)
@click.pass_context
def _ontologies(ctx, consortium, workspace, details):
    """Find ontologies fields in workspace and write summary to stdout."""
    G = _recursive_default_dict()
    for ontology_fields in ontologies(ctx.obj['output_path'], consortium, workspace, details):
        for ontology_field in ontology_fields:
            G[ontology_field['consortium_name']][ontology_field['workspace_name']][ontology_field['entity_name']] = ontology_fields['ontology_fields']
    print(json.dumps(G))


@utility.command(name='qa')
@click.option('--consortium', default=None, help='Filter, only this consortium.')
@click.option('--workspace', default=None, help='Filter, only this workspace.')
@click.option('--details', default=False, help='Include error details.', show_default=True, is_flag=True)
@click.pass_context
def qa(ctx, consortium, workspace, details):
    """Report on analysis."""
    summaries = []
    with open(f"{ctx.obj['output_path']}/analysis.ndjson") as input_stream:
        for line in input_stream.readlines():
            summaries.append(json.loads(line))
    logger.warning("TODO - write to dataframe.")


@utility.command(name='bucket')
@click.option('--source',
              type=click.Choice(['FHIR', 'DB'], case_sensitive=False),
              help="Copy FHIR directories or working databases to bucket"
              )
@click.option('--bucket_name',
              help="Google bucket to write to."
              )
@click.pass_context
def _bucket(ctx, source, bucket_name):
    """Backup files to terra workspace bucket."""
    logger.info(f"TODO: {ctx.command.name} copy from {ctx.obj['output_path']} to bucket")


@utility.command(name='sample')
@click.option('--consortium', default=None, help='Filter, only this consortium.')
@click.option('--workspace', default=None, help='Filter, only this workspace.')
@click.option('--depth', default=10, help='Number of patients.', show_default=True)
@click.pass_context
def _sample(ctx, consortium, workspace, depth):
    """Export source data for <depth> patients to file system."""
    if workspace:
        if not consortium:
            consortium = _consortium_from_workspace(ctx.obj['config'], workspace)
        workspace_names = [(consortium, workspace)]
    else:
        workspace_names = fetch_workspace_names(ctx.obj['output_path'], requested_consortium_name=consortium,
                                                workspace_name=workspace)

    dir_name = f"{ctx.obj['output_path']}/sample"

    drs_reader = DRSReader(ctx.obj['output_path'])

    from random import seed
    from random import randint
    seed()

    def _generate_samples(_consortium_name, _workspace_name):
        for _consortium_name, _workspace in normalize(
                ctx.obj['output_path'],
                requested_consortium_name=_consortium_name,
                workspace_name=_workspace_name,
                validate_buckets=False,
                config=ctx.obj['config']
        ):
            for index, patient_id in zip(range(depth), _workspace.patients):
                patient = _workspace.patients[patient_id]
                yield patient
                # if 'family_relationship_values' in patient:
                #     if patient['family_relationship_values'] != [{}]:
                #         for family_relationship_value in patient['family_relationship_values']:
                #             yield {
                #                 'entityType': 'family_relationship',
                #                 'name': str(randint(1, 10000000)),
                #                 'attributes': family_relationship_value
                #             }
                if 'family' in patient:
                    yield patient['family']

                for specimen in patient.get('specimens', []):
                    yield specimen
                    for task in specimen.get('tasks', []):
                        if task['entityType'] != 'Task':
                            #  don't yield _implied tasks
                            yield task
                        for source in task['outputs']:
                            for field in task['outputs'][source]:
                                # remove the drs_uri
                                if 'drs_uri' in task['outputs'][source][field]:
                                    del task['outputs'][source][field]['drs_uri']
                                yield {
                                    'entityType': 'blob',
                                    'attributes': task['outputs'][source][field]
                                }
                                drs = drs_reader.get(task['outputs'][source][field]['url'].split('/')[-1])
                                if drs:
                                    yield {
                                        'entityType': 'drs',
                                        'attributes': drs
                                    }

    for consortium_name, workspace_name in workspace_names:

        emitters = {}

        for entity in _generate_samples(consortium_name, workspace_name):
            subdir = f"{dir_name}/{consortium_name}/{workspace_name}"
            file_path = f"{subdir}/{entity['entityType']}.ndjson"

            emitter = emitters.get(file_path, None)
            if emitter is None:
                os.makedirs(subdir, exist_ok=True)
                emitter = open(file_path, "w")
                logger.info(f"Writing {file_path}")
                emitters[file_path] = emitter

            if entity['entityType'] not in ['blob', 'drs']:
                json.dump({'attributes': entity['attributes'],
                           'name': entity['name'],
                           'entityType': entity['entityType']},
                          emitter,
                          separators=(',', ':'),
                          )
            else:
                json.dump(entity['attributes'],
                          emitter,
                          separators=(',', ':'),
                          )

            emitter.write('\n')

        for emitter in emitters.values():
            emitter.close()



"""Utility methods."""
import re
import uuid

from anvil.terra.workspace import Workspace


def make_identifier(*args):
    """Create legal fhir id."""
    return re.sub(r"[^A-Za-z0-9\-\.]", "-", ".".join(str(a) for a in args))[-64:]


def join(*args):
    """Create pipe separated string."""
    return "|".join(str(a) for a in args)


def make_id(workspace_name, resource_id):
    """Create legal fhir id, a reproducible SHA-1 hash of a workspace_name and resource_id."""
    namespace = uuid.UUID(bytes=bytes(bytearray(("xxxxxxxxxxxxxxxx" + workspace_name)[-16:], 'utf-8')))
    return str(uuid.uuid5(namespace, resource_id))


def make_workspace_id(resource):
    """Deduce workspace id."""
    workspace_name = None
    if issubclass(resource.__class__, Workspace):
        workspace_name = resource.attributes.workspace.name
    elif hasattr(resource, 'workspace_name'):
        workspace_name = resource.workspace_name
    assert workspace_name, f"missing workspace_name {resource}"
    return make_identifier(workspace_name)


# note coordinate with StructureDefintions /fhir/config.yaml::canonical
CANONICAL = "https://nih-ncpi.github.io/ncpi-fhir-ig"



"""Represent fhir entity."""

from anvil.transformers.fhir import CANONICAL, join
from anvil.transformers.fhir.practitioner import Practitioner
from anvil.transformers.fhir.organization import Organization

import logging
logger = logging.getLogger(__name__)
from anvil.transformers.fhir.disease_normalizer import text_ontology, ontology_text, disease_system
from anvil.transformers.fhir import make_workspace_id


class ResearchStudy:
    """Create fhir entity."""

    class_name = "research_study"
    resource_type = "ResearchStudy"

    @staticmethod
    def slug(resource):
        """Make id."""
        return make_workspace_id(resource)

    @staticmethod
    def build_entity(workspace):
        """Create fhir entity."""
        study_id = ResearchStudy.slug(workspace)
        workspace_name = workspace.attributes.workspace.name
        investigator_name = workspace.investigator
        if not investigator_name:
            logger.warning(f'{study_id} missing investigator')
        diseaseOntologyId = workspace.diseaseOntologyId

        diseaseOntologyIdText = 'Missing'
        if diseaseOntologyId and diseaseOntologyId not in ontology_text:
            logger.error(f'{study_id} missing >{diseaseOntologyId}< see anvil.transformers.fhir.disease_normalizer.disease_text')
            if ':' not in diseaseOntologyId:
                logger.info(f"study_id {study_id} diseaseOntologyId {diseaseOntologyId} does not look like an ontology id")
                diseaseOntologyId = None
        else:
            diseaseOntologyIdText = ontology_text.get(diseaseOntologyId, 'Missing')

        # workspace = workspace.attributes.workspace.attributes
        institute = workspace.institute
        if not institute:
            logger.warning(f'{study_id} missing institute')
        study_name = study_id
        attribution = study_id
        short_name = study_id
        key = study_id
        condition = None
        if diseaseOntologyId:
            prefix, code = diseaseOntologyId.split(':')
            if prefix not in disease_system:
                logger.warning(f'{study_id} diseaseOntologyId not configured {prefix}')
            else:    
                condition = [
                    {
                        "coding": [
                            {
                                "system": disease_system[prefix],
                                "code": code,
                                "display": diseaseOntologyIdText,
                            }
                        ]
                    }
                ]

        entity = {
            "resourceType": ResearchStudy.resource_type,
            "id": study_id,
            "meta": {
                "profile": [
                    "http://hl7.org/fhir/StructureDefinition/ResearchStudy"
                ]
            },
            "identifier": [
                {
                    "system": "https://anvil.terra.bio/#workspaces/anvil-datastorage/",
                    "value": workspace_name,
                },
                {
                    "system": "urn:ncpi:unique-string",
                    "value": join(ResearchStudy.resource_type, key),
                },
            ],
            "extension": [
                {
                    "url": f"{CANONICAL}/StructureDefinition/related-organization",
                    "extension": [
                        {
                            "url": "organization",
                            "valueReference": {
                                "reference": f"Organization/{Organization.slug(workspace)}"
                            },
                        }
                    ],
                }
            ],
            "title": study_name,
            "status": "completed",
            "sponsor": {
                "reference": f"Organization/{study_id.lower()}"
            }
        }

        if investigator_name:
            entity["principalInvestigator"] = {
                "reference": f"Practitioner/{Practitioner.slug(workspace)}"
            }

        if attribution:
            entity["identifier"].append({"value": attribution})

        if short_name:
            entity["extension"].append(
                {
                    "url": f"{CANONICAL}/StructureDefinition/display-name",
                    "valueString": short_name,
                }
            )
        if condition:
            entity['condition'] = condition

        return entity



# -*- coding: utf-8 -*-
"""Google gcloud access_token handling class for smart-on-fhir/client-py FHIR client."""
import logging
from subprocess import Popen, PIPE

from fhirclient import auth

logger = logging.getLogger(__name__)

REGISTERED = []


class FHIRAuthError(Exception):
    """Reports any problem retrieving access token from gcloud."""

    pass


class GoogleFHIRAuth(auth.FHIRAuth):
    """Google gcloud access_token handling class for smart-on-fhir/client-py FHIR client.

    Requires:
        `pip install -e git+https://github.com/smart-on-fhir/client-py#egg=fhirclient`

    :param access_token: Optional access token, if none provided `gcloud auth print-access-token` is used

    Examples:
        from fhirclient import client
        from anvil.clients.smart_auth import GoogleFHIRAuth
        settings = {
            'app_id': 'my_web_app',
            'api_base': 'https://healthcare.googleapis.com/v1/projects/gcp-testing-308520/locations/us-east4/datasets/testset/fhirStores/fhirstore/fhir'
        }
        smart = client.FHIRClient(settings=settings)
        # optionally pass token
        smart.server.auth = GoogleFHIRAuth(access_token=''ya29.abcd...')
        smart.prepare()
        assert smart.ready, "server should be ready"
        # search for all ResearchStudy
        import fhirclient.models.researchstudy as rs
        [s.title for s in rs.ResearchStudy.where(struct={}).perform_resources(smart.server)]
        >>>
        ['1000g-high-coverage-2019', 'my NCPI research study example']
    """

    auth_type = 'bearer'

    def __init__(self, state=None, access_token=None):
        """Initialize access_token, call super."""
        self.access_token = access_token
        if not self.access_token:
            self.access_token = self._get_auth_value()
        super(GoogleFHIRAuth, self).__init__(state=state)

    @property
    def ready(self):
        """Return True if access_token exists."""
        return True if self.access_token else False

    def reset(self):
        """Clear access_token."""
        super(GoogleFHIRAuth, self).reset()
        self.access_token = None

    def can_sign_headers(self):
        """Return True if access_token exists."""
        return True if self.access_token is not None else False

    def signed_headers(self, headers):
        """Return updated HTTP request headers, if possible, raises if there is no access_token."""
        if not self.can_sign_headers():
            raise Exception("Cannot sign headers since I have no access token")

        if headers is None:
            headers = {}
        headers['Authorization'] = "Bearer {0}".format(self.access_token)

        return headers

    def reauthorize(self, server):
        """Perform reauthorization.

        Args:
            - server - The Server instance to use

        Returns:
            - output - The launch context dictionary, or None on failure
        """
        logger.debug("SMART AUTH: Refreshing token")
        self.access_token = self._get_auth_value()
        return {'access_token': self._access_token}

    def handle_401(self, response, **kwargs):
        """Handle failed requests when authorization failed.

        This gets called after a failed request when an HTTP 401 error
        occurs. This then tries to refresh the access token in the event
        that it expired.

        Args:
            request (object): The failed request object

        """
        if not response.status_code == 401 and not response.status_code == 403:
            return response

        # Free the original connection
        response.content
        response.close()

        # copy the request to resend
        newreq = response.request.copy()

        self.access_token = None
        logger.debug("handle_401, cleared _access_token, retrying with new token")

        self.access_token = self._get_auth_value()
        newreq.headers["Authorization"] = "Bearer {0}".format(self.access_token)

        _response = response.connection.send(newreq, **kwargs)
        _response.history.append(response)
        _response.request = newreq

        return _response

    @property
    def state(self):
        """Save state."""
        s = super(GoogleFHIRAuth, self).state
        if self.access_token is not None:
            s['access_token'] = self.access_token

        return s

    def from_state(self, state):
        """Update ivars from given state information."""
        super(GoogleFHIRAuth, self).from_state(state)
        self.access_token = state.get('access_token') or self.access_token

    def _get_auth_value(self):
        """Return the Authorization header value for the request.

        This gets called when added the Authorization header to the request.
        This fetches the access token from the refresh token if the access token is missing.

        """
        if not self.access_token:
            try:
                # get the local access token using gcloud
                cmd = ['gcloud', 'auth', 'print-access-token']
                logger.debug(f"getting gcloud_access_token {cmd}")
                p = Popen(cmd, stdout=PIPE, stderr=PIPE)
                gcloud_access_token, stderr = p.communicate()
                # remove CR
                gcloud_access_token = gcloud_access_token.decode("utf-8").rstrip()
                assert len(gcloud_access_token) > 0, f'get gcloud_access_token MUST have an access token {stderr}'
                logger.debug(f"gcloud_access_token {gcloud_access_token}")
                self.access_token = gcloud_access_token
                if logger.level == logging.DEBUG:
                    logger.debug(self.access_token)
                    print(self.access_token)
                    # expires_at = datetime.fromisoformat(self.access_token['expires_at'])
                    # logger.debug(f'access token expires in {str(expires_at - now)}')
                    # # add padding
                    # logger.debug(base64.b64decode(self._access_token.split('.')[1] + "==="))
            except Exception as e:
                raise FHIRAuthError(f"Failed to get gcloud_access_token {e}")
        return self.access_token


# register class
if GoogleFHIRAuth not in REGISTERED:
    print('GoogleFHIRAuth registered')
    GoogleFHIRAuth.register()
    REGISTERED.append(GoogleFHIRAuth)


class KidsFirstFHIRAuth(auth.FHIRAuth):
    """KidsFirst handling class for smart-on-fhir/client-py FHIR client.

    Requires:
        `pip install -e git+https://github.com/smart-on-fhir/client-py#egg=fhirclient`

    :param cookie: Mandatory AWSELBAuthSessionCookie token.

    Examples:
        cookie = 'AWSELBAuthSessionCookie-0=ABC123abc...'

        from fhirclient import client
        from anvil.clients.smart_auth import KidsFirstFHIRAuth
        settings = {
            'app_id': 'my_web_app',
            'api_base': 'https://kf-api-fhir-service.kidsfirstdrc.org'
        }
        kids_first = client.FHIRClient(settings=settings)
        # optionally pass token
        kids_first.server.auth = KidsFirstFHIRAuth(cookie=cookie)
        kids_first.prepare()
        assert kids_first.ready, "server should be ready"
        # search for all ResearchStudy
        import fhirclient.models.researchstudy as rs
        [s.title for s in rs.ResearchStudy.where(struct={}).perform_resources(kids_first.server)]
        >>>
        ['OpenDIPG: ICR London', 'Genomic Studies of Orofacial Cleft Birth Defects', 'Genome-wide Sequencing to Identify the Genes Responsible for Enchondromatoses and Related Malignant Tumors', 'Kids First: Craniofacial Microsomia: Genetic Causes and Pathway Discovery', 'Whole Genome Sequencing of African and Asian Orofacial Clefts Case-Parent Triads', 'Kids First: Genetics of Structural Defects of the Kidney and Urinary Tract', 'National Heart, Lung, and Blood Institute (NHLBI) Bench to Bassinet Program: The Gabriella Miller Kids First Pediatric Research Program of the Pediatric Cardiac Genetics Consortium (PCGC)', 'Kids First: Genetics at the Intersection of Childhood Cancer and Birth Defects', 'TARGET: Neuroblastoma (NBL)', 'Genomic Analysis of Familial Leukemia', 'Kids First: Genomics of Orofacial Cleft Birth Defects in Latin American Families', 'Genomics of Orthopaedic Disease Program', 'Kids First: The Genetics of Microtia in Hispanic Populations', 'Gabriella Miller Kids First Pediatric Research Program in Novel Cancer Susceptibility in Families (from BASIC3)', 'Genomic Analysis of Esophageal Atresia and Tracheoesophageal Fistulas and Associated Congenital Anomalies', 'Discovering the Genetic Basis of Human Neuroblastoma: A Gabriella Miller Kids First Pediatric Research Program (Kids First) Project', 'Kids First: Genomic Analysis of a Cohort with Infantile Hemangiomas Associated with Multi-organ Structural Birth Defects', 'Pharmacokinetics, Pharmacodynamics, and Safety Profile of Understudied Drugs Administered to Children per Standard of Care (POP02)', "Pediatric Brain Tumor Atlas - Children's Brain Tumor Tissue Consortium", 'BCH Structural Birth Defects Collaboration: Syndromic cranial dysinnervation disorders', 'Expanded Ewing sarcoma cohort for tumor genomics and association with DNA repair deficiencies, clinical presentation, and outcome', 'Genetic Basis of Disorders/Differences of Sex Development (DSD)', 'Pediatric Brain Tumor Atlas: PNOC', 'Kids First: Whole genome sequencing of nonsyndromic craniosynostosis', 'Kids First: Genomic Analysis of Congenital Heart Defects and Acute Lymphoblastic Leukemia in Children with Down Syndrome', 'TARGET: Acute Myeloid Leukemia (AML)', 'Kids First: Germline and Somatic Variants in Myeloid Malignancies in Children', 'Genomic Analysis of Congenital Diaphragmatic Hernia', 'An Integrated Clinical and Genomic Analysis of Treatment Failure in Pediatric Osteosarcoma']
    """

    auth_type = 'cookie'

    def __init__(self, state=None, cookie=None):
        """Initialize access_token, call super."""
        self.cookie = cookie
        super(KidsFirstFHIRAuth, self).__init__(state=state)

    @property
    def ready(self):
        """Return True if access_token exists."""
        return True if self.cookie else False

    def reset(self):
        """Clear access_token."""
        super(KidsFirstFHIRAuth, self).reset()
        self.cookie = None

    def can_sign_headers(self):
        """Return True if access_token exists."""
        return True if self.cookie is not None else False

    def signed_headers(self, headers):
        """Return updated HTTP request headers, if possible, raises if there is no access_token."""
        if not self.can_sign_headers():
            raise Exception("Cannot sign headers since I have no cookie")

        if headers is None:
            headers = {}
        headers['cookie'] = self.cookie
        logger.debug(headers)
        return headers

    def handle_callback(self, url, server):
        """Return the launch context."""
        logger.debug('handle_callback')
        raise Exception(f"{self} cannot handle callback URL")

    def reauthorize(self, server):
        """Perform reauthorization.

        Args:
            - server - The Server instance to use

        Returns:
            - output - The launch context dictionary, or None on failure
        """
        logger.debug("SMART AUTH: Refreshing token")
        return {'cookie': self.cookie}

    def handle_401(self, response, **kwargs):
        """Handle failed requests when authorization failed.

        This gets called after a failed request when an HTTP 401 error
        occurs. This then tries to refresh the access token in the event
        that it expired.

        Args:
            request (object): The failed request object

        """
        for keyword in ['authorize', 'login']:
            if keyword in response.url:
                raise Exception("The token is no longer valid, please re-harvest from KidsFirst browser.")
        if not response.status_code == 401 and not response.status_code == 403:
            return response
        raise Exception("The user represented by this cookie does not have access.")

    @property
    def state(self):
        """Save state."""
        s = super(KidsFirstFHIRAuth, self).state
        if self.cookie is not None:
            s['cookie'] = self.cookie

        return s

    def from_state(self, state):
        """Update ivars from given state information."""
        super(KidsFirstFHIRAuth, self).from_state(state)
        self.cookie = state.get('cookie') or self.cookie


# register class
if KidsFirstFHIRAuth not in REGISTERED:
    print('KidsFirstFHIRAuth registered')
    KidsFirstFHIRAuth.register()
    REGISTERED.append(KidsFirstFHIRAuth)



import os
from collections import defaultdict

import click
import logging
import json

import requests

from anvil.etl.transform import _consortium_from_workspace
from anvil.etl.transformers.fhir_writer import ensure_data_store_name
from anvil.etl.transformers.normalizer import get_pickled_workspace, fetch_workspace_names
from anvil.etl.utilities.shell_helper import run_cmd

logger = logging.getLogger(__name__)


@click.group(name='fhir')
@click.pass_context
def fhir(ctx):
    pass  # dummy


# @fhir.command(name='enable-ig')
# @click.option('--token',
#               default=None,
#               help='gcloud access token. if null cmd will exec `gcloud auth application-default print-access-token`',
#               show_default=True)
# @click.pass_context
# def _enable_ig(ctx, token):
#     """Enable IG in a datastore"""
#     if not token:
#         token = run_cmd("gcloud auth application-default print-access-token")
#     output_path = ctx.obj["output_path"]
#     workspace_mapping = _extract_workspace_mapping(output_path)
#     headers = {
#         "Authorization": f"Bearer {token}",
#         "Content-Type": "application/json; charset=utf-8"
#     }
#     project = os.environ['GOOGLE_PROJECT']
#     location = os.environ['GOOGLE_LOCATION']
#     data_set = os.environ['GOOGLE_DATASET']
#     for data_store in workspace_mapping['data_store']:
#         url = f"https://healthcare.googleapis.com/v1beta1/projects/{project}/locations/{location}/datasets/{data_set}/fhirStores/{data_store}?updateMask=validationConfig"
#         response = requests.patch(
#             url=url,
#             headers=headers,
#             json={
#                 "validationConfig": {
#                     "enabledImplementationGuides": ["https://ncpi-fhir.github.io/ncpi-fhir-ig/ImplementationGuide/NCPI-FHIR-Implementation-Guide"]
#                 }
#             }
#         )
#         if response.status_code == 200:
#             logger.info(f"IG enabled on {data_store}")
#         else:
#             logger.info(("could.not.enable.ig", data_store, response.status_code, response.text))
#


# @fhir.command(name='create-data-stores')
# @click.pass_context
# def create_data_stores(ctx):
#     """Create data_stores, enable IG."""
#     # TODO - filter consortium, workspace, data-store
#     output_path = ctx.obj["output_path"]
#     workspace_mapping = _extract_workspace_mapping(output_path)
#     list_cmd = f'gcloud beta healthcare fhir-stores list --dataset=$GOOGLE_DATASET  --location=$GOOGLE_LOCATION  --format="table[no-heading](ID)"'
#     data_store_list = run_cmd(list_cmd)
#     cmds = [f"gcloud beta healthcare fhir-stores create {data_store} --dataset=$GOOGLE_DATASET --location=$GOOGLE_LOCATION --version=R4 --enable-update-create" for data_store in workspace_mapping['data_store'].keys() if data_store not in data_store_list]
#     if len(cmds) > 0:
#         run_cmd("\n".join(cmds))
#     data_store_list = run_cmd(list_cmd)
#     error = False
#     for data_store in workspace_mapping['data_store']:
#         if data_store not in data_store_list:
#             logger.error(f"{data_store} not found in:\n{data_store_list}")
#             error = True
#     if not error:
#         logger.info("All data stores exist.")
#     cmds = [f"gcloud beta healthcare fhir-stores import gcs  {data_store}  --dataset=$GOOGLE_DATASET  --location=$GOOGLE_LOCATION --gcs-uri=gs://$GOOGLE_BUCKET/IG/*.json --content-structure=resource-pretty --async" for data_store in workspace_mapping['data_store'].keys() if data_store in data_store_list]
#     if len(cmds) > 0:
#         print(run_cmd("\n".join(cmds)))
#     logger.info("IG loaded on all.")
#     _log_console_link()


# @fhir.command(name='drop-data-stores')
# @click.pass_context
# def drop_data_stores(ctx):
#     """Create data_stores, enable IG."""
#     # TODO - filter consortium, workspace, data-store
#     output_path = ctx.obj["output_path"]
#     workplace_mapping = _extract_workspace_mapping(output_path)
#     list_cmd = f'gcloud beta healthcare fhir-stores list --dataset=$GOOGLE_DATASET  --location=$GOOGLE_LOCATION  --format="table[no-heading](ID)"'
#     data_store_list = run_cmd(list_cmd)
#     cmds = [f"gcloud beta healthcare fhir-stores delete {data_store} --dataset=$GOOGLE_DATASET --location=$GOOGLE_LOCATION --quiet" for data_store in workplace_mapping['data_store'].keys() if data_store in data_store_list]
#     if len(cmds) > 0:
#         run_cmd("\n".join(cmds))
#     data_store_list = run_cmd(list_cmd)
#     for data_store in workplace_mapping['data_store']:
#         if data_store in data_store_list:
#             logger.error(f"{data_store} delete failed, found in:\n{data_store_list}")
#     logger.info("All data stores deleted.")

#
# @fhir.command(name='load-data-stores')
# @click.pass_context
# def load_data_stores(ctx):
#     """Load uploaded data into data stores."""
#     output_path = ctx.obj["output_path"]
#     workplace_mapping = _extract_workspace_mapping(output_path)
#     workspace_datastore_mapping = workplace_mapping['data_store']
#
#     cmds = []
#     for data_store, workspaces in workspace_datastore_mapping.items():
#         for mapping in workspaces:
#             object_path = f"fhir/{mapping['consortium']}/{mapping['name']}"
#             for subdir in ['public', 'protected']:
#                 cmds.append(f"gcloud healthcare fhir-stores import gcs {data_store} --location=$GOOGLE_LOCATION --dataset=$GOOGLE_DATASET --content-structure=resource --async --gcs-uri=gs://$GOOGLE_BUCKET/{object_path}/{subdir}/*.ndjson")
#     run_cmd("\n".join(cmds))
#     _log_console_link()
#
#
# @fhir.command(name='upload-data')
# @click.option('--consortium', default=None, help='Filter, only this consortium.')
# @click.option('--workspace', default=None, help='Filter, only this workspace')
# @click.option('--drop', default=True, help='Delete FHIR daa in bucket first')
# @click.pass_context
# def upload_data(ctx, consortium, workspace, drop):
#     """Copy IG to bucket."""
#     output_path = ctx.obj["output_path"]
#     workspace_consortium_mapping = _extract_workspace_mapping(output_path)['consortium']
#     copy_commands = []
#     check_commands = []
#     uploaded_workspaces = []
#     for consortium_name, workspaces in workspace_consortium_mapping.items():
#         for mapping in workspaces:
#             if consortium and consortium != consortium_name:
#                 continue
#             if workspace and workspace != mapping['name']:
#                 continue
#             assert 'name' in mapping, mapping
#             dir_name = f"{output_path}/fhir/{consortium_name}/{mapping['name']}"
#             obj_name = f"fhir/{consortium_name}/{mapping['name']}"
#             assert os.path.isdir(dir_name), f"{dir_name} does not exist?"
#             # gzip, recursive, no clobber
#             if drop:
#                 copy_commands.append(f'gsutil -m rm -r gs://{os.environ["GOOGLE_BUCKET"]}/{obj_name}/public')
#                 copy_commands.append(f'gsutil -m rm -r gs://{os.environ["GOOGLE_BUCKET"]}/{obj_name}/protected')
#             copy_commands.append(f'gsutil -m cp -J -r -n  {dir_name}/public/** gs://{os.environ["GOOGLE_BUCKET"]}/{obj_name}/public')
#             copy_commands.append(f'gsutil -m cp -J -r -n  {dir_name}/protected/** gs://{os.environ["GOOGLE_BUCKET"]}/{obj_name}/protected')
#             check_commands.append(f'gsutil ls gs://{os.environ["GOOGLE_BUCKET"]}/{obj_name}')
#             uploaded_workspaces.append((consortium_name, mapping['name']))
#     run_cmd('\n'.join(copy_commands))
#     results = run_cmd('\n'.join(check_commands))
#     for consortium_name, workspace_name in uploaded_workspaces:
#         assert f"{consortium_name}/{workspace_name}/public" in results, ('missing.obj', f"{consortium_name}/{workspace_name}/public", results)
#         assert f"{consortium_name}/{workspace_name}/protected" in results, ('missing.obj', f"{consortium_name}/{workspace_name}/protected", results)
#
#





#!/usr/bin/env python3

"""Read meta data from terra workspaces."""

import logging
import re
from attrdict import AttrDict
import firecloud.api as FAPI
import json
import click
import os
from urllib.parse import urlparse
from collections import defaultdict

from anvil.etl.utilities.entities import Entities

logging.basicConfig(level=logging.INFO, format='%(asctime)s %(filename)s %(levelname)-8s %(message)s')
logger = logging.getLogger('anvil.etl.extractors.terra')


# where all workspaces are kept w/in terra
DEFAULT_NAMESPACE = 'anvil-datastorage'

# workspace patterns
DEFAULT_CONSORTIUMS = (
    ('CMG', 'AnVIL_CMG_.*'),
    ('CCDG', 'AnVIL_CCDG_.*'),
    ('GTEx', '^AnVIL_GTEx_V8_hg38$'),
    ('Public', '^1000G-high-coverage-2019$'),
    ('NHGRI', '^AnVIL_NHGRI'),
    ('NIMH', '^AnVIL_NIMH'),
)


def recursive_default_dict():
    """Recursive default dict, any key defaults to a dict."""
    return defaultdict(recursive_default_dict)


def get_workspaces(namespaces=None, name_pattern=None):
    """Filter terra workspaces by namespaces and project_pattern.

    Args:
        namespaces ([str]): Optional, list of workspace `namespace` to match ex: 'anvil-datastorage'.
        name_pattern (str): Optional, regexp to match workspace `name` ex: 'AnVIL_CCDG.*'.

    Returns:
        dict: keys ['accessLevel', 'public', 'workspace', 'workspaceSubmissionStats']

    """
    logger.debug(f"get_entities {namespaces} {name_pattern}")

    workspaces = FAPI.list_workspaces()
    workspaces = workspaces.json()

    if namespaces:
        workspaces = [AttrDict(w) for w in workspaces if w['workspace']['namespace'] in namespaces]

    if name_pattern:
        workspaces = [AttrDict(w) for w in workspaces if re.match(name_pattern, w['workspace']['name'], re.IGNORECASE)]

    # normalize fields
    for w in workspaces:
        if 'project_files' not in w.workspace:
            w.workspace.project_files = []
    return workspaces


def get_entities(namespace='anvil-datastorage', workspace=None, entity_name=None):
    """Return all entities in a workspace."""
    logger.debug(f"get_entities {namespace} {workspace} {entity_name}")
    try:
        entities = [AttrDict(e) for e in FAPI.get_entities(namespace, workspace, entity_name).json()]
        return entities
    except Exception as e:
        logger.error(f"{workspace} {entity_name} {e}")
        return []


def get_schema(namespace, workspace):
    """Fetch all entity types."""
    logger.debug(f"get_schema {namespace} {workspace}")
    schema = FAPI.list_entity_types(namespace=namespace, workspace=workspace).json()
    if 'statusCode' in schema['schema']:
        raise Exception(schema)
    return schema


def extract_bucket_fields(output_path=None, entities=None, workspace_name=None):
    """Query db, sample first row in all entities, determine fields that have references to bucket object."""
    if not entities:
        assert output_path
        entities = Entities(path=f"{output_path}/terra_entities.sqlite")
    for workspace in entities.get_by_label('workspace'):
        workspace = AttrDict(workspace)
        if workspace_name and workspace.workspace.name != workspace_name:
            continue
        schema = AttrDict(entities.get_edges_by_label(src=workspace.workspace.name, src_name='workspace', dst_name='schema'))
        for entity_name in schema.schema.keys():
            if entity_name == 'schema':
                continue
            child = entities.get_edges_by_label(src=workspace.workspace.name, src_name='workspace', dst_name=entity_name, limit=1)
            if entity_name not in child:
                logger.error(('invalid.child', workspace.consortium_name, workspace.workspace.name, entity_name, child))
                continue
            bucket_keys = [k for k, v in child[entity_name]['attributes'].items() if isinstance(v, str) and v.startswith('gs://')]
            if len(bucket_keys) == 0:
                continue
            buckets = [child[entity_name]['attributes'][key] for key in bucket_keys]
            buckets = list(set([urlparse(b).netloc for b in buckets]))
            # TODO - move this exception to config or other ?
            if entity_name == 'sequencing' and 'AnVIL_CMG_Broad' in workspace.workspace.name:
                bucket_keys = 'crai_or_bai_path,crai_path,cram_or_bam_path,cram_path,md5_path'.split(',')
            yield {
                'consortium_name': workspace.consortium_name,
                'workspace_name': workspace.workspace.name,
                'entity_name': entity_name,
                'bucket_fields': bucket_keys,
                'buckets': buckets
            }


def make_entity_key(workspace_name, entity):
    """Edges and vertices are stored with this key structure."""
    return f"{entity['entityType']}/{workspace_name}/{entity['name']}"


@click.group()
@click.pass_context
def cli(ctx):
    """Read meta data from terra workspaces."""
    # ensure that ctx.obj exists and is a dict
    # in case we want to eventually chain these commands together into a pipeline
    assert ctx.obj['output_path'], "Missing DEFAULT_OUTPUT_PATH"


@cli.command('clean')
@click.pass_context
def clean(ctx):
    """Remove database."""
    path = f"{ctx.obj['output_path']}/terra_entities.sqlite"
    try:
        os.remove(path)
        logger.info(('removed', path))
    except OSError as e:
        logger.warning((e, path))


@cli.command('extract')
@click.option('--namespace', default=DEFAULT_NAMESPACE, help=f'Terra namespace default={DEFAULT_NAMESPACE}')
@click.pass_context
def extract_workspaces(ctx, namespace):
    """Read workspaces from terra, write to database. Do this first! May take several minutes."""
    output_path = ctx.obj['output_path']
    consortiums = ctx.obj['config']['consortiums']    
    logger.info(f"Extracting metadata for {len(consortiums)} consortiums, this may take several minutes.")
    entities = Entities(path=f"{output_path}/terra_entities.sqlite")
    for consortium_name, config in consortiums.items():
        projects = get_workspaces(namespace, name_pattern=config['workspaces'])
        for workspace in projects:
            logger.info((consortium_name, workspace.workspace.name))
            workspace.consortium_name = consortium_name
            schema = FAPI.list_entity_types(namespace=namespace, workspace=workspace.workspace.name).json()
            if 'statusCode' in schema:
                logger.error(('no.schema', consortium_name, workspace.workspace.name, schema))
                continue
            entities.put(key=workspace.workspace.name, label='workspace', data=workspace)
            entities.put(key=f"schema/{workspace.workspace.name}", label='schema', data=schema)
            entities.put_edge(
                src=workspace.workspace.name, src_name='workspace',
                dst=f"schema/{workspace.workspace.name}", dst_name='schema',
            )
            for entity_name in schema.keys():
                problem_occurred = False
                for entity in FAPI.get_entities(namespace, workspace.workspace.name, entity_name).json():
                    if problem_occurred:
                        break
                    if not isinstance(entity, dict):
                        # write a problem record
                        entities.put(key=f"problem/{workspace.workspace.name}/{entity_name}", label='problem', data={'error': f'no.entity.returned {entity_name}'})
                        entities.put_edge(
                            src=workspace.workspace.name, src_name='workspace',
                            dst=f"problem/{workspace.workspace.name}/{entity_name}", dst_name='problem',
                        )
                        problem_occurred = True
                        continue
                    # entity['name'] aka 'id'
                    id = make_entity_key(workspace.workspace.name, entity)
                    entities.put(key=id, label=entity_name, data=entity)
                    entities.put_edge(
                        src=workspace.workspace.name, src_name='workspace',
                        dst=id, dst_name=entity_name
                    )
    entities.commit(True)
    entities.index()


@cli.command('cat')
@click.option('--workspace', default=None, help='<Regexp> e.g "AnVIL_CCDG.*"')
@click.pass_context
def cat(ctx, workspace):
    """Read workspaces and their children from database, write to stdout."""
    output_path = ctx.obj['output_path']
    name_pattern = workspace
    entities = Entities(path=f"{output_path}/terra_entities.sqlite")
    for workspace in entities.get_by_label('workspace'):
        if name_pattern:
            if not re.match(name_pattern, workspace['workspace']['name'], re.IGNORECASE):
                continue
        workspace = AttrDict(workspace)
        children = entities.get_edges(src=workspace.workspace.name, src_name='workspace')
        workspace.children = children
        print(json.dumps(workspace))


@cli.command('bucket_fields')
@click.option('--workspace', default=None, help='filter, only this workspace')
@click.pass_context
def bucket_fields(ctx, workspace):
    """Read workspaces from database, determine children with bucket references, write to stdout."""
    output_path = ctx.obj['output_path']
    print(json.dumps([bucket_fields for bucket_fields in extract_bucket_fields(output_path, workspace_name=workspace)]))


# @cli.command('bucket_graph')
# @click.option('--details', default=False, is_flag=True, help="Show the details.")
# @click.pass_context
# def bucket_graph(ctx, details):
#     """Read bucket fields from database, analyze patterns, write to stdout."""
#     output_path = ctx.obj['output_path']

#     G = recursive_default_dict()
#     for bf in extract_bucket_fields(output_path):
#         # {
#         #     'consortium_name': workspace.consortium_name,
#         #     'workspace_name': workspace.workspace.name,
#         #     'entity_name': entity_name,
#         #     'bucket_fields': bucket_keys,
#         #     'buckets': buckets
#         # }
#         bf = AttrDict(bf)
#         G[bf.consortium_name][bf.entity_name][','.join(sorted(bf.bucket_fields))]['workspaces'][bf.workspace_name] = bf.buckets

#     analysis = recursive_default_dict()
#     for consortium_name in G:
#         for entity_name in G[consortium_name]:
#             version_count = len(G[consortium_name][entity_name].keys())
#             analysis[consortium_name][entity_name]['version_count'] = version_count
#             if details:
#                 analysis[consortium_name][entity_name]['details'] = G[consortium_name][entity_name]
#     print(json.dumps(analysis))


@cli.command('schema')
@click.option('--workspace', default=None, help='filter, only this workspace')
@click.pass_context
def schema(ctx, workspace):
    """Read schema fields from database, analyze patterns, write to stdout."""
    output_path = ctx.obj['output_path']
    workspace_name = workspace
    entities = Entities(path=f"{output_path}/terra_entities.sqlite")

    G = recursive_default_dict()
    for workspace in entities.get_by_label('workspace'):
        workspace = AttrDict(workspace)
        if workspace_name and workspace.workspace.name != workspace_name:
            continue
        schema = entities.get_edges_by_label(src=workspace.workspace.name, src_name='workspace', dst_name='schema')
        for entity_name in schema['schema']:
            assert 'attributeNames' in schema['schema'][entity_name], (entity_name, schema)
            attributeNames = ','.join(sorted(schema['schema'][entity_name]['attributeNames']))
            G[workspace.consortium_name][entity_name][attributeNames][workspace.workspace.name] = {}

    analysis = recursive_default_dict()
    for consortium_name in G:
        for entity_name in G[consortium_name]:
            version_count = len(G[consortium_name][entity_name].keys())
            analysis[consortium_name][entity_name]['version_count'] = version_count
            analysis[consortium_name][entity_name]['details'] = G[consortium_name][entity_name]
    print(json.dumps(analysis))



"""Represent fhir entity."""

from anvil.transformers.fhir import make_id
from anvil.transformers.fhir.organization import Organization
from urllib.parse import urlsplit, urlunsplit

from anvil.transformers.fhir.patient import Patient


def strip_port(url):
    """Remove port from url (BUG in gen3)."""
    if not url:
        return url
    parts = urlsplit(url)
    return urlunsplit((parts.scheme, parts.netloc.split(':')[0], parts.path, parts.query, parts.fragment,))


class GENOMIC_FILE:
    """Define constants."""

    class AVAILABILITY:
        """Define constants."""

        IMMEDIATE = "Immediate Download"
        COLD_STORAGE = "Cold Storage"

    class DATA_TYPE:
        """Define constants."""

        ALIGNED_READS = "Aligned Reads"
        ALIGNED_READS_INDEX = "Aligned Reads Index"
        EXPRESSION = "Expression"
        GVCF = "gVCF"
        GVCF_INDEX = "gVCF Index"
        HISTOLOGY_IMAGES = "Histology Images"
        NUCLEOTIDE_VARIATION = "Simple Nucleotide Variations"
        OPERATION_REPORTS = "Operation Reports"
        PATHOLOGY_REPORTS = "Pathology Reports"
        RADIOLOGY_IMAGES = "Radiology Images"
        RADIOLOGY_REPORTS = "Radiology Reports"
        UNALIGNED_READS = "Unaligned Reads"
        VARIANT_CALLS = "Variant Calls"
        VARIANT_CALLS_INDEX = "Variant Calls Index"
        ANNOTATED_SOMATIC_MUTATIONS = "Annotated Somatic Mutations"
        GENE_EXPRESSION = "Gene Expression"
        GENE_FUSIONS = "Gene Fusions"
        ISOFORM_EXPRESSION = "Isoform Expression"
        SOMATIC_COPY_NUMBER_VARIATIONS = "Somatic Copy Number Variations"
        SOMATIC_STRUCTURAL_VARIATIONS = "Somatic Structural Variations"

    class FORMAT:
        """Define constants."""

        BAI = "bai"
        BAM = "bam"
        CRAI = "crai"
        CRAM = "cram"
        DCM = "dcm"
        FASTQ = "fastq"
        GPR = "gpr"
        GVCF = "gvcf"
        IDAT = "idat"
        PDF = "pdf"
        SVS = "svs"
        TBI = "tbi"
        VCF = "vcf"


# http://fhir.kids-first.io/ValueSet/data-type
data_type_dict = {
    GENOMIC_FILE.DATA_TYPE.ALIGNED_READS: {
        "coding": [
            {
                "system": "http://fhir.kids-first.io/CodeSystem/data-type",
                "code": "C164052",
                "display": "Aligned Sequence Read",
            }
        ],
        "text": GENOMIC_FILE.DATA_TYPE.ALIGNED_READS,
    },
    GENOMIC_FILE.DATA_TYPE.GENE_EXPRESSION: {
        "coding": [
            {
                "system": "http://fhir.kids-first.io/CodeSystem/data-type",
                "code": "C16608",
                "display": "Gene Expression",
            }
        ],
        "text": GENOMIC_FILE.DATA_TYPE.GENE_EXPRESSION,
    },
    GENOMIC_FILE.DATA_TYPE.GENE_FUSIONS: {
        "coding": [
            {
                "system": "http://fhir.kids-first.io/CodeSystem/data-type",
                "code": "C20195",
                "display": "Gene Fusion",
            }
        ],
        "text": GENOMIC_FILE.DATA_TYPE.GENE_FUSIONS,
    },
    GENOMIC_FILE.DATA_TYPE.OPERATION_REPORTS: {
        "coding": [
            {
                "system": "http://fhir.kids-first.io/CodeSystem/data-type",
                "code": "C114420",
                "display": "Operative Report",
            }
        ],
        "text": GENOMIC_FILE.DATA_TYPE.OPERATION_REPORTS,
    },
    GENOMIC_FILE.DATA_TYPE.PATHOLOGY_REPORTS: {
        "coding": [
            {
                "system": "http://fhir.kids-first.io/CodeSystem/data-type",
                "code": "C28277",
                "display": "Pathology Report",
            }
        ],
        "text": GENOMIC_FILE.DATA_TYPE.PATHOLOGY_REPORTS,
    },
    GENOMIC_FILE.DATA_TYPE.ANNOTATED_SOMATIC_MUTATIONS: {
        "coding": [
            {
                "system": "http://fhir.kids-first.io/CodeSystem/data-type",
                "code": "C18060",
                "display": "Somatic Mutation",
            }
        ],
        "text": GENOMIC_FILE.DATA_TYPE.ANNOTATED_SOMATIC_MUTATIONS,
    },
    GENOMIC_FILE.DATA_TYPE.UNALIGNED_READS: {
        "coding": [
            {
                "system": "http://fhir.kids-first.io/CodeSystem/data-type",
                "code": "C164053",
                "display": "Unaligned Sequence Read",
            }
        ],
        "text": GENOMIC_FILE.DATA_TYPE.UNALIGNED_READS,
    },
}


class DocumentReference:
    """Render entity."""

    class_name = "document_reference"
    resource_type = "DocumentReference"

    @staticmethod
    def slug(blob):
        """Make id."""
        return make_id(blob.sample.workspace_name, blob.attributes.name)

    @staticmethod
    def build_entity(blob, subject):
        """Render entity."""
        study_id = blob.sample.workspace_name
        subject_slug = Patient.slug(subject)
        blob_slug = DocumentReference.slug(blob)
        # acl = None
        # size = blob.attributes.size
        # url_list = [blob.attributes.name]
        file_name = blob.attributes.name.split('/')[-1]
        file_format = file_name.split('.')[-1]
        # data_type = file_format
        # time_created = blob.attributes.time_created

        entity = {
            "resourceType": DocumentReference.resource_type,
            "id": blob_slug,
            "meta": {
                "profile": [
                    "https://ncpi-fhir.github.io/ncpi-fhir-ig/StructureDefinition/ncpi-research-document-reference"
                ]
            },
            "identifier": [
                {
                    "system": f"https://anvil.terra.bio/#workspaces/anvil-datastorage/{study_id}",
                    "value": blob.attributes.name,
                },
                {
                    "system": "urn:ncpi:unique-string",
                    "value": blob_slug,
                },
            ],
            "status": "current",
            "custodian": {
                "reference": f"Organization/{Organization.slug(subject)}"
            }
        }

        if subject_slug:
            entity["subject"] = {
                "reference": f"Patient/{subject_slug}"
            }

        content = {}

        # start attachment
        if 'ga4gh_drs_uri' in blob.attributes:
            entity["meta"] = {
                "profile": [
                    "http://fhir.ncpi-project-forge.io/StructureDefinition/ncpi-drs-document-reference"
                ]
            }
        url = strip_port(blob.attributes.get('ga4gh_drs_uri', None))
        if not url:
            url = blob.attributes['name']
        content["attachment"] = {
            "url": url
        }

        # end attachment

        if file_format:
            content["format"] = {
                "display": file_format
            }

        if content:
            entity.setdefault("content", []).append(content)

        return entity



"""Represent fhir entity."""

from anvil.transformers.fhir import make_id
from anvil.transformers.fhir.research_study import ResearchStudy
from anvil.terra.subject import Subject


class Patient:
    """Create fhir entity."""

    class_name = "patient"
    resource_type = "Patient"

    @staticmethod
    def slug(subject):
        """Make id."""
        return make_id(subject.workspace_name, subject.id)

    @staticmethod
    def build_entity(subject):
        """Create fhir entity."""
        assert issubclass(subject.__class__, Subject), f"{subject}"

        subject_id_slug = Patient.slug(subject)

        # ethnicity = None
        # race = None
        # species = None
        # gender = None

        entity = {
            "resourceType": Patient.resource_type,
            "id": subject_id_slug,
            "meta": {
                "profile": [
                    "http://hl7.org/fhir/StructureDefinition/Patient"
                ]
            },
            "identifier": [
                {
                    "system": f"https://anvil.terra.bio/#workspaces/anvil-datastorage/{subject.workspace_name}",
                    "value": subject.id,
                },
                {
                    "system": "urn:ncpi:unique-string",
                    "value": f"{subject.workspace_name}/Patient/{subject.id}",
                },
            ],
            "managingOrganization": {
                "reference": f"Organization/{ResearchStudy.slug(subject)}"
            }
        }

        # ethnicity QUESTION: http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity
        # if ethnicity:
        #     if omb_ethnicity_category.get(ethnicity):
        #         entity.setdefault("extension", []).append(
        #             {
        #                 "url": "http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity",
        #                 "extension": [
        #                     omb_ethnicity_category[ethnicity],
        #                     {"url": "text", "valueString": ethnicity},
        #                 ],
        #             }
        #         )

        # race QUESTION: http://hl7.org/fhir/us/core/StructureDefinition/us-core-race
        #     if race:
        #         if omb_race_category.get(race):
        #             entity.setdefault("extension", []).append(
        #                 {
        #                     "url": "http://hl7.org/fhir/us/core/StructureDefinition/us-core-race",
        #                     "extension": [
        #                         omb_race_category[race],
        #                         {"url": "text", "valueString": race},
        #                     ],
        #                 }
        #             )

        # species QUESTION: species ?
        #     if species:
        #         if species_dict.get(species):
        #             entity.setdefault("extension", []).append(species_dict[species])

        if subject.gender:
            entity["gender"] = subject.gender

        return entity



import logging
from anvil.etl.utilities.entities import Entities
from google.cloud.storage import Client
from .terra import extract_bucket_fields


logger = logging.getLogger(__name__)


# @cli.command('clean')
# @click.option('--output_path', default=DEFAULT_OUTPUT_PATH, help=f'output path default={DEFAULT_OUTPUT_PATH}')
# def clean(output_path):
#     """Remove database."""
#     path = f"{output_path}/google_entities.sqlite"
#     try:
#         os.remove(path)
#         logger.info(('removed', path))
#     except OSError as e:
#         logger.warning((e, path))


def extract_buckets(output_path, user_project):
    """Get all gs:// bucket references in all workspaces from terra, retrieve blob info from google, write to db."""
    entities = Entities(path=f"{output_path}/google_entities.sqlite")
    logger.info(('user_project', user_project))
    client = Client(project=user_project)
    bucket_fields = [bf for bf in extract_bucket_fields(output_path)]
    already_done = set()
    for bucket_field in bucket_fields:
        for bucket in bucket_field['buckets']:
            if bucket in already_done:
                continue
            logger.info((bucket_field['consortium_name'], bucket_field['workspace_name'], bucket))
            already_done.add(bucket)
            blobs = client.list_blobs(bucket)
            for blob in blobs:
                _properties = dict(blob._properties)
                _properties['path'] = blob.path
                _properties['public_url'] = blob.public_url
                _properties['url'] = f"gs://{_properties['bucket']}/{_properties['name']}"
                entities.put(key=_properties['url'], label='Blob', data=_properties)
            entities.commit()
    entities.index()



"""Wraps  firecloud api."""

from anvil.util.cache import memoize
import firecloud.api as FAPI
import logging
import re
from attrdict import AttrDict

logger = logging.getLogger('anvil.terra.api')

USER_PROJECT = None


def whoami():
    """Wrap fapi's whoami.

    Returns:
        str: google id

    """
    return FAPI.whoami()


@memoize
def get_projects(namespaces=None, project_pattern=None):
    """Filter terra workspaces by namespaces and project_pattern.

    Args:
        namespaces ([str]): Optional, list of workspace `namespace` to match ex: 'anvil-datastorage'.
        project_pattern (str): Optional, regexp to match workspace `name` ex: 'AnVIL_CCDG.*'.

    Returns:
        dict: keys ['accessLevel', 'public', 'workspace', 'workspaceSubmissionStats']

    """
    logger.debug(f"get_entities {namespaces} {project_pattern}")

    workspaces = FAPI.list_workspaces()
    workspaces = workspaces.json()
    logger.debug(f'all workspaces len {len(workspaces)}')
    logger.debug(f"all namespaces {set(sorted([w['workspace']['namespace'] for w in workspaces]))}")

    if namespaces:
        workspaces = [AttrDict(w) for w in workspaces if w['workspace']['namespace'] in namespaces]

    logger.debug(f'after namespace >{namespaces}<', len(workspaces))
    logger.debug(f"all namespaces {set(sorted([w['workspace']['namespace'] for w in workspaces]))}")

    if project_pattern:
        workspaces = [AttrDict(w) for w in workspaces if re.match(project_pattern, w['workspace']['name'], re.IGNORECASE)]

    logger.debug(f'after project_pattern >{project_pattern}<', len(workspaces))
    logger.debug(f"all workspaces {set(sorted([w['workspace']['name'] for w in workspaces]))}")

    # normalize fields
    for w in workspaces:
        if 'project_files' not in w.workspace:
            w.workspace.project_files = []
    return workspaces


@memoize
def get_entities(namespace='anvil-datastorage', workspace=None, entity_name=None):
    """Return all entities in a workspace."""
    logger.debug(f"get_entities {namespace} {workspace} {entity_name}")
    try:
        entities = [AttrDict(e) for e in FAPI.get_entities(namespace, workspace, entity_name).json()]
        return entities
    except Exception as e:
        logger.error(f"{workspace} {entity_name} {e}")
        return []
    


@memoize
def get_schema(namespace, workspace):
    """Fetch all entity types."""
    logger.debug(f"get_schema {namespace} {workspace}")
    return FAPI.list_entity_types(namespace=namespace, workspace=workspace).json()
